/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include <stdio.h>
#include <stdlib.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <Util/Time.h>
#include <Util/cmdLineParser.h>
#include <Util/Half/Half.h>
#include "LaplacianMatrix/SocketedMultigrid/MultigridProcess.h"
#include "LaplacianMatrix/SocketedMultigrid/MultigridServer.h"
#include "LaplacianMatrix/SocketedMultigrid/SocketData.h"
#include "Util/ImageStream.h"

#define DEFAULT_BUFLEN 1024

typedef unsigned __int16 LType;

cmdLineInt Port( "port" , 0 );
cmdLineString Address( "address" ) , Prefix( "prefix" ) , TempDir( "temp" );
cmdLineString Storage( "storage" );
cmdLineString Pixels( "pixels" ) , Labels( "labels" ) , Out( "out" );
cmdLineInt Threads( "threads" , 1 ) , Priority( "priority" , THREAD_PRIORITY_NORMAL );
cmdLineReadable InCore( "inCore" );
cmdLineInt Index( "index" , 0 );
cmdLineReadable* params[]=
{
	&Pixels , &Labels , &Out , &Port , &Address , &Prefix , &Index , &TempDir , &Storage , &Priority ,
	&Threads , &InCore
};
void ShowUsage(char* ex)
{
	printf( "Usage %s:\n",ex) , fflush( stdout );
	printf( "\t--%s <composited image pixels>\n" , Pixels.name ) , fflush( stdout );
	printf( "\t--%s <image labels>\n" , Labels.name ) , fflush( stdout );
	printf( "\t--%s <output image>\n" , Out.name ) , fflush( stdout );
	printf( "\t--%s <server connection address>\n" , Address.name ) , fflush( stdout );
	printf( "\t--%s <server port>\n" , Port.name , Port.value) , fflush( stdout );
	printf( "\t[--%s <client index>=%d]\n" , Index.name , Index.value ) , fflush( stdout );
	printf( "\t[--%s <scratch directory>]\n" , TempDir.name ) , fflush( stdout );
	printf( "\t[--%s <number of computation threads>=%d]\n" , Threads.name , Threads.value ) , fflush( stdout );
	printf( "\t[--%s]\n" , InCore.name ) , fflush( stdout );

#if 0
	printf( "\t[--%s <preferred address prefix>]\n" , Prefix.name ) , fflush( stdout );
	printf("\t[--%s <thread priority>=%d]\n" , Priority.name , Priority.value ) , fflush( stdout );
	printf( "\t[--%s <temporary storage type>]\n" , Storage.name ) , fflush( stdout );
#endif
}
template< class StorageType >
int Execute( int argc , char* argv[]  )
{
	MultiStreamIOServer ioServer;

	int w1 , h1;
	GetReadSize< half >( Pixels.value , w1 , h1 );
	printf( "Input image size: %d x %d\n" , w1 , h1 );

	SocketedSuperMultigridClient< 3 , StorageType , half , LType > client;
	if( !client.SetUp( Address.value , Prefix.value , Port.value  ,Index.value , w1 , h1 , Threads.value , InCore.set , &ioServer ) )
	{
		fprintf(stderr,"SocketedMultigridClient::SetUp failed\n");
		return EXIT_FAILURE;
	}
	StreamingGrid *pixels = NULL , *labels = NULL , *out = NULL;
	if( !debugFlags.noInput ) pixels = GetReadStream< half >( Pixels.value , w1 , h1 , false , &ioServer , Threads.value , Priority.value );
	else					  pixels = new NULLStreamingGrid( w1 * 3  * sizeof( half ) , h1 );
	int w2,h2;
	if( !client.sharpen() )
	{
		if( !Labels.set )
		{
			ShowUsage(argv[0]);
			return EXIT_FAILURE;
		}
		if( !debugFlags.noInput )
			labels = GetReadStream< LType >( Labels.value , w2 , h2 , false , & ioServer , Threads.value , Priority.value );
		else
		{
			GetReadSize< half >( Labels.value , w2 , h2 );
			labels = new NULLStreamingGrid( sizeof( LType ) * w2 * 3 , h2 );
		}

		if(w1!=w2 || h1!=h2)
		{
			fprintf(stderr,"Image and label dimensions differ: %d x %d != %d x %d\n",w1,h1,w2,h2);
			return EXIT_FAILURE;
		}
	}

	DefaultOutputTileWidth = client.tileWidth( );
	DefaultOutputTileHeight = client.tileHeight( );
	if( client.tileExtension() ) DefaultOutputTileExtension = client.tileExtension( );

	if( !debugFlags.noOutput )
		out = new WImageClipper<float>( Out.value , client.size() , client.height() , true , 0 , 0 , client.cSize() , client.cHeight() , client.quality() , &ioServer , client.hdr() , Threads.value , Priority.value );
	else
		out = new NULLStreamingGrid( client.cSize() * sizeof( float ) * 3 , client.cHeight() );

	char tmpDir[1024];
	if(TempDir.set)	sprintf( tmpDir , "TMP=%s" , TempDir.value );
	else			sprintf( tmpDir , "TMP=" );
	_putenv(tmpDir);
	double t=Time();
	if( client.sharpen() )	client.Run( pixels , NULL   , out );
	else					client.Run( pixels , labels , out );
	size_t current,peak;
	WorkingSetInfo(current,peak);
	printf("Running Time: %f\n",Time()-t) , fflush( stdout );
	printf("Peak working set: %d MB\n",peak>>20) , fflush( stdout );
	printf("Packets Sent / Received: %lld / %lld\n",packetsSent,packetsReceived) , fflush( stdout );

	if(pixels) delete pixels;
	if(labels) delete labels;
	if(out) delete out;
	MyWinSock::UnLoad();
	return EXIT_SUCCESS;
}
int main( int argc , char* argv[])
{
	_setmaxstdio(1024);
	MyWinSock::Load();
	int paramNum=sizeof(params)/sizeof(cmdLineReadable*);
	if( !cmdLineParse( argc-1 , &argv[1] , paramNum , params , true ) ) return EXIT_FAILURE;

	if( !Pixels.set || !Out.set || !Address.set )
	{
		ShowUsage(argv[0]);
		return EXIT_FAILURE;
	}
	if( !Port.set )
	{
		int a1 , a2 , a3 , a4;
		if( sscanf( Address.value , " %d.%d.%d.%d:%d " , &a1 , &a2 , &a3 , &a4 , &Port.value ) != 5 )
		{
			ShowUsage(argv[0]);
			return EXIT_FAILURE;
		}
		sprintf( Address.value, "%d.%d.%d.%d" , a1 , a2 , a3 , a4 );
	}

	if( Storage.set )
		if		( !strcasecmp( Storage.value , "float" ) )	return Execute< half >	( argc , argv );
		else if	( !strcasecmp( Storage.value , "double" ) )	return Execute< double >( argc , argv );
		else if	( !strcasecmp( Storage.value , "half" ) )	return Execute< half >	( argc , argv );
		else
		{
			fprintf( stderr , "Unrecognized storage type: %s\n" ,  Storage.value );
			return EXIT_FAILURE;
		}
	else return Execute< half >( argc , argv );
}
