/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include <winsock2.h>
#include <ws2tcpip.h>
#include "Util/MultiStreamIO.h"
#include "Util/Socket.h"
#include "Util/Half/half.h"


template<int Type>
Vector<double> Multiply(const SparseMatrix<double>& A,const Vector<double>& x);

template< int Type >
Vector< double > Multiply( const SparseMatrix< double >& A , const Vector< double >& x )
{
	return A*x;
}
template<> inline Vector<double> Multiply<ZERO_DERIVATIVE>(const SparseMatrix<double>& A,const Vector<double>& x)
{
	Vector<double> y = A*x;
	double average=0;
	for(size_t i=0;i<x.Dimensions();i++)	average+=x[i];
	average/=x.Dimensions();
	for(size_t i=0;i<x.Dimensions();i++)	y[i]+=average;
	return y;
}
template<> inline Vector<double> Multiply<PERIODIC_BOUNDARY>(const SparseMatrix<double>& A,const Vector<double>& x)
{
	Vector<double> y = A*x;
	double average=0;
	for(size_t i=0;i<x.Dimensions();i++)	average+=x[i];
	average/=x.Dimensions();
	for(size_t i=0;i<x.Dimensions();i++)	y[i]+=average;
	return y;
}
template<class Type>
Vector<double> Multiply(const SparseMatrix<double>& A,const Vector<double>& x)	{	return A*x;	}
template<int Type>
int MySolveConjugateGradient(const SparseMatrix<double>& A,const Vector<double>& b,const int& iters,Vector<double>& x)
{
	double eps=1e-16;
	Vector<double> r = b - A*x;
	Vector<double> d = r;
	double delta_new = r.Dot(r);
	double delta_0 = delta_new;
	int i;
	for(i=0; i<iters && delta_new>eps*delta_0 ;i++)
	{
		Vector<double> q = Multiply<Type>(A,d);
		double alpha = delta_new / d.Dot(q);
		x = x + d*alpha;
		if( !(i%50) )	r = b - Multiply<Type>(A,x);
		else			r = r - q*alpha;

		double delta_old = delta_new;
		delta_new = r.Dot(r);
		double beta = delta_new / delta_old;
		d = r + d*beta;
	}
	return i;
}

void SetPaddedSize( int width , int height , int minRes , int& paddedWidth , int& paddedHeight )
{
	int blockSize;
	long long _paddedWidth  = minRes;
	long long _paddedHeight = minRes;
	int domainW = FiniteElements1D<float,Type,Degree>::DomainSize( width );
	int domainH = FiniteElements1D<float,Type,Degree>::DomainSize( height );
	while( _paddedWidth <domainW )	_paddedWidth  *= 2;
	while( _paddedHeight<domainH )	_paddedHeight *= 2;

	blockSize = int( _paddedWidth/minRes );
	blockSize *= 2;
	while( _paddedWidth-blockSize>domainW ) _paddedWidth -= blockSize;

	blockSize = int( _paddedHeight/minRes );
	blockSize *= 2;
	while( _paddedHeight-blockSize>domainH ) _paddedHeight -= blockSize;

	_paddedWidth  = FiniteElements1D<float,Type,Degree>::Dimension( int( _paddedWidth  ) );
	_paddedHeight = FiniteElements1D<float,Type,Degree>::Dimension( int( _paddedHeight ) );

	paddedWidth  = int(_paddedWidth);
	paddedHeight = int(_paddedHeight);
}
bool IsDownSamplable( int width , int height , int iters , int& newWidth , int& newHeight )
{
	if( (width&3) || (height&3) || (width<20) )	return false;
	newWidth  = width>>1;
	newHeight = height>>1;
	return true;
}
bool IsDownSamplable(int width,int height,int& newWidth,int& newHeight)
{
	if( (width&3) || (height&3) || (width<20) )	return false;
	newWidth  = width>>1;
	newHeight = height>>1;
	return true;
}
bool IsDownSamplable(int dCount,ClientSocket* cSockets,int sockCount)
{
	for(int i=0;i<sockCount;i++)
		if ( (cSockets[i].clientData.start>>dCount)&3 || ((cSockets[i].clientData.end-cSockets[i].clientData.start)>>dCount)<20)	return false;
	return true;
}
template< class Real >
void UpSampleRectangularSolution( const Vector< Real >&low , int width ,int height , Vector< Real >& high )
{
	int width2 , height2;
	FiniteElements1D< double , Type , Degree >::FullProlongationStencil wStencil , hStencil;
	FiniteElements1D< double , Type , Degree >::ProlongationStencil( width  , wStencil , width2 );
	FiniteElements1D< double , Type , Degree >::ProlongationStencil( height , hStencil , height2 );
	high.Resize( width2 * height2 );
	for( int i=0 ; i<width ; i++ )
	{
		int iStart = FiniteElements1D< double , Type , Degree >::FullProlongationStencil::ProlongationStencil::Start( i );
		int ii;
		if( i<Degree )				ii = i;
		else if( i>=width-Degree)	ii = 2*Degree+(i-(width-1));
		else						ii = Degree;
		for( int j=0 ; j<height ; j++ )
		{
			int jj;
			if( j<Degree )				jj = j;
			else if( j>=height-Degree)	jj = 2*Degree+(j-(height-1));
			else						jj = Degree;
			int jStart = FiniteElements1D< double , Type , Degree >::FullProlongationStencil::ProlongationStencil::Start( j );
			for( int x=0 ; x<FiniteElements1D< double , Type , Degree >::FullProlongationStencil::ProlongationStencil::Size ; x++ )
				if( iStart+x<0 || iStart+x>=width2 ) continue;
				else
					for( int y=0 ; y<FiniteElements1D< double , Type , Degree >::FullProlongationStencil::ProlongationStencil::Size ; y++ )
						if( jStart+y<0 || jStart+y>=height2 ) continue;
						else 
							high[ (jStart+y)*width2 + (iStart+x) ] += low[j*width+i] * wStencil.caseTable[ii].values[x] * hStencil.caseTable[jj].values[y];
		}
	}
}
template<class Real>
void SetRectangularLaplacianMatrix( SparseMatrix< Real > & lap , int width , int height , double iWeight=0 )
{
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::FullDotProductStencil dotStencil1,d2DotStencil1,dotStencil2,d2DotStencil2;
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(width,dotStencil1,0,0);
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(width,d2DotStencil1,1,1,false);
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(height,dotStencil2,0,0);
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(height,d2DotStencil2,1,1,false);
	SetRectangularLaplacianMatrix( dotStencil1 , d2DotStencil1 , dotStencil2 , d2DotStencil2 , lap , width , height , iWeight );
}
template<class Real>
void SetRectangularLaplacianMatrix( DotProductStencil& dotMajor , DotProductStencil& d2DotMajor , DotProductStencil& dotMinor , DotProductStencil& d2DotMinor ,
								    SparseMatrix< Real > & lap , int width , int height , double iWeight=0 )
{
	lap.Resize( width*height );

	int ii , jj;
	int xStart , xEnd , yStart , yEnd;
	for(int y=0;y<height;y++)
	{
		if		(y<Degree)			jj = y							, yStart = -y		, yEnd = Degree;
		else if	(y>height-1-Degree)	jj = 2*Degree+(y-(height-1))	, yStart = -Degree	, yEnd = height-1-y;
		else						jj = Degree						, yStart = -Degree	, yEnd = Degree;

		for(int x=0;x<width;x++)
		{
			if		(x<Degree)			ii = x						, xStart = -x		, xEnd = Degree;
			else if	(x>width-1-Degree)	ii = 2*Degree+(x-(width-1))	, xStart = -Degree	, xEnd = width-1-x;
			else						ii = Degree					, xStart = -Degree	, xEnd = Degree;

			int idx = x+y*width;
			lap.SetGroupSize( idx , (yEnd-yStart+1)*(xEnd-xStart+1) );
			int _i = 0;
			for(int yy=yStart;yy<=yEnd;yy++)
				for(int xx=xStart;xx<=xEnd;xx++)
				{
					lap[idx][_i  ].N=(x+xx)+(y+yy)*width;
					lap[idx][_i++].Value=
						dotMajor.caseTable[ii].values[xx+Degree]*d2DotMinor.caseTable[jj].values[yy+Degree]+
						d2DotMajor.caseTable[ii].values[xx+Degree]*dotMinor.caseTable[jj].values[yy+Degree]+
						(dotMajor.caseTable[ii].values[xx+Degree]*dotMinor.caseTable[jj].values[yy+Degree])*iWeight;
				}
		}
	}
}

template<class Real>
void SetSphericalLaplacianMatrix(SparseMatrix<Real>& lap,int width,int height)
{
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::FullDotProductStencil dotStencil1,d2DotStencil1,dotStencil2,d2DotStencil2;
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(width,dotStencil1,0,0);
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(width,d2DotStencil1,1,1,false);
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(height,dotStencil2,0,0);
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(height,d2DotStencil2,1,1,false);
	SetSphericalLaplacianMatrix( dotStencil1 , d2DotStencil1 , dotStencil2 , d2DotStencil2 , lap , width , height );
}
template<class Real>
void SetSphericalLaplacianMatrix(DotProductStencil& dotMajor,DotProductStencil& d2DotMajor,DotProductStencil& dotMinor,DotProductStencil& d2DotMinor,
								 SparseMatrix<Real>& lap,int width,int height)
{
	lap.Resize(width*height);

	for(int x=0;x<width;x++)
		for(int y=0;y<height;y++)
		{
			int idx1 = x + y*width;
			lap.SetGroupSize( idx1 , (2*Degree+1)*(2*Degree+1) );
			for(int i=-Degree;i<=Degree;i++)
				for(int j=-Degree;j<=Degree;j++)
				{
					int idx2 = (i+Degree)+(j+Degree)*(2*Degree+1);
					int xx=(x+i+width)%width;
					int yy= y+j;
					if(yy<0)
					{
						yy = -yy-1;
						if(xx<width/2)	xx = width/2-1-xx;
						else			xx = width/2+width-1-xx;
					}
					else if(yy>=height)
					{
						yy = height-1-(yy-height);
						xx = width-1-xx;
					}
					lap[idx1][idx2].N = xx+yy*width;
					lap[idx1][idx2].Value = 
						dotMajor.caseTable[Degree].values[i+Degree]*d2DotMinor.caseTable[Degree].values[j+Degree]+
						d2DotMajor.caseTable[Degree].values[i+Degree]*dotMinor.caseTable[Degree].values[j+Degree];
				}
		}
}
template< class Real >
void UpSampleSphericalSolution( const Vector< Real >&low , int width , int height , Vector< Real >& high )
{
	int width2 , height2;
	FiniteElements1D< double , Type , Degree >::FullProlongationStencil wStencil , hStencil;
	FiniteElements1D< double , Type , Degree >::ProlongationStencil( width  , wStencil , width2  );
	FiniteElements1D< double , Type , Degree >::ProlongationStencil( height , hStencil , height2 );
	high.Resize( width2 * height2 );
	for( int i=0 ; i<width ; i++ )
	{
		int iStart = FiniteElements1D< double , Type , Degree >::FullProlongationStencil::ProlongationStencil::Start( i );
		for( int j=0 ; j<height ; j++ )
		{
			int jStart = FiniteElements1D< double , Type , Degree >::FullProlongationStencil::ProlongationStencil::Start( j );
			for( int x=0 ; x<FiniteElements1D< double , Type , Degree >::FullProlongationStencil::ProlongationStencil::Size ; x++ )
			{
				for( int y=0 ; y<FiniteElements1D< double , Type , Degree >::FullProlongationStencil::ProlongationStencil::Size ; y++ )
				{
					int xx = ( iStart + x + width2 ) % width2;
					int yy= jStart+y;
					if( yy<0 )
					{
						yy = -yy-1;
						if( xx<width2/2 )	xx = width2/2          - 1 - xx;
						else				xx = width2/2 + width2 - 1 - xx;
					}
					else if( yy>=height2 )
					{
						yy = height2-1-(yy-height2);
						xx = width2-1-xx;
					}
					high[ yy*width2 + xx ] += low[j*width+i] * wStencil.caseTable[Degree].values[x] * hStencil.caseTable[Degree].values[y];
				}
			}
		}
	}
}

template<class Real>
void SetSphericalLaplacianMatrix(SparseMatrix<Real>& lap,int width,int height,double iWeight)
{
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::FullDotProductStencil dotStencil1,d2DotStencil1,dotStencil2,d2DotStencil2;
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(width,dotStencil1,0,0);
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(width,d2DotStencil1,1,1,false);
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(height,dotStencil2,0,0);
	FiniteElements1D<double,Type,Degree>::DotProduct<Type,Degree>::DotProductStencil(height,d2DotStencil2,1,1,false);
	SetSphericalLaplacianMatrix( dotStencil1 , d2DotStencil1 , dotStencil2 , d2DotStencil2 , lap , width , height , iWeight );
}
template<class Real>
void SetSphericalLaplacianMatrix(DotProductStencil& dotMajor,DotProductStencil& d2DotMajor,DotProductStencil& dotMinor,DotProductStencil& d2DotMinor,
								 SparseMatrix<Real>& lap,int width,int height,double iWeight)
{
	lap.Resize(width*height);

	for(int x=0;x<width;x++)
		for(int y=0;y<height;y++)
		{
			int idx1 = x + y*width;
			lap.SetGroupSize( idx1 , (2*Degree+1)*(2*Degree+1) );
			for(int i=-Degree;i<=Degree;i++)
				for(int j=-Degree;j<=Degree;j++)
				{
					int idx2 = (i+Degree)+(j+Degree)*(2*Degree+1);
					int xx=(x+i+width)%width;
					int yy= y+j;
					if(yy<0)
					{
						yy = -yy-1;
						if(xx<width/2)	xx = width/2-1-xx;
						else			xx = width/2+width-1-xx;
					}
					else if(yy>=height)
					{
						yy = height-1-(yy-height);
						xx = width-1-xx;
					}
					lap[idx1][idx2].N = xx+yy*width;
					lap[idx1][idx2].Value = 
						dotMajor.caseTable[Degree].values[i+Degree]*d2DotMinor.caseTable[Degree].values[j+Degree]+
						d2DotMajor.caseTable[Degree].values[i+Degree]*dotMinor.caseTable[Degree].values[j+Degree]+
						(dotMajor.caseTable[Degree].values[i+Degree]*dotMinor.caseTable[Degree].values[j+Degree])*iWeight;
				}
		}
}
///////////////////////////
// SphericalSynchronizer //
///////////////////////////
template<int Channels>
SphericalSynchronizer<Channels>::SphericalSynchronizer(void)
{
	_cCount = 0;
	_clientSockets = NULL;
}
template<int Channels>
void SphericalSynchronizer<Channels>::init( int width , const int* widths , int cCount , int rPasses , int pPasses , int vCycles )
{
	_width=width;
	_cCount=cCount;
	_rPasses=rPasses;
	_pPasses=pPasses;
	_vCycles=vCycles;

	if( _clientSockets ) delete[] _clientSockets;

	_clientSockets = new ClientSocket[cCount];
	if(!_clientSockets)
	{
		fprintf(stderr,"Failed to allocate SphericalSynchronizer::_clientSockets\n");
		exit(0);
	}
	for( int i=0 ; i<cCount ; i++ )
	{
		_clientSockets[i].client = GetListenSocket(_clientSockets[i].port);
		if (_clientSockets[i].client == INVALID_SOCKET)	fprintf(stderr,"Failed to set server in SphericalSynchronizer\n") , exit(0);
		if(i)	_clientSockets[i].clientData.start = _clientSockets[i-1].clientData.end;
		else	_clientSockets[i].clientData.start = 0;
		_clientSockets[i].clientData.end = _clientSockets[i].clientData.start+widths[i];
	}
}
template<int Channels>
SphericalSynchronizer<Channels>::~SphericalSynchronizer(void)
{
	delete[] _clientSockets;
	_clientSockets=NULL;
}
template<int Channels>	int SphericalSynchronizer<Channels>::clients(void)		const	{ return _cCount; }
template<int Channels>	int SphericalSynchronizer<Channels>::port(int cIndex)	const	{ return _clientSockets[cIndex].port; }
template<int Channels>
template<class DataType>
void SphericalSynchronizer<Channels>::Run(void)
{
	for(int i=0;i<_cCount;i++)
	{
		_clientSockets[i].client = AcceptSocket( _clientSockets[i].client );
		if ( _clientSockets[i].client == INVALID_SOCKET ) exit(0);
	}

	Pointer( DataType ) buffer1;
	Pointer( DataType ) buffer2;
	buffer1 = AllocArray< DataType >( _width * Channels , 1 , "SphericalSynchronizer< Channels >::Run (buffer1)" );
	buffer2 = AllocArray< DataType >( _width * Channels , 1 , "SphericalSynchronizer< Channels >::Run (buffer2)" );
	for ( int v = 0 ; v < _vCycles ; v++ )
	{
		//////////////////////////////////////////
		// First synchronize on the restriction //
		//////////////////////////////////////////

		// Synchronize the head of the stream
		for(int k=0;k<_rPasses;k++)
		{
			for(int i=0;i<_cCount;i++)
			{
				int w = _clientSockets[i].clientData.end-_clientSockets[i].clientData.start;
				ReceiveOnSocket( _clientSockets[i].client , buffer1 , sizeof(DataType)*w*Channels , "Failed to get restriction head from client" );
				for( int x = 0 ; x < w ; x++ )
					for( int c = 0; c < Channels ; c++ )
						buffer2[(x+_clientSockets[i].clientData.start)*Channels+c] = buffer1[w*c+x];
			}
			for(int x=0;x<_width/2;x++)
			{
				int xx = _width/2-1-x;
				for( int c=0 ; c<Channels ; c++ ) buffer1[ xx*Channels+c ] = buffer2[x*Channels+c];
			}
			for(int x=_width/2;x<_width;x++)
			{
				int xx = _width/2+_width-1-x;
				for( int c=0 ; c<Channels ; c++ ) buffer1[ xx*Channels+c ] = buffer2[x*Channels+c];
			}
			for(int i=0;i<_cCount;i++)
			{
				int w = _clientSockets[i].clientData.end-_clientSockets[i].clientData.start;
				for( int x = 0 ; x < w ; x++ )
					for( int c = 0 ; c < Channels ; c++ )
						buffer2[w*c+x] = buffer1[(x+_clientSockets[i].clientData.start)*Channels+c];
				SendOnSocket( _clientSockets[i].client , buffer2 , sizeof(DataType)*w*Channels , "Failed to send restriction head to client" ) ;
			}
		}
		// Synchronize the tail of the stream
		for(int k=0;k<_rPasses;k++)
		{
			for(int i=0;i<_cCount;i++)
			{
				int w = _clientSockets[i].clientData.end-_clientSockets[i].clientData.start;
				ReceiveOnSocket( _clientSockets[i].client , buffer1 , sizeof(DataType)*w*Channels , "Failed to get restriction tail from client" );
				for( int x = 0 ; x < w ; x++ )
					for( int c = 0; c < Channels ; c++ )
						buffer2[(x+_clientSockets[i].clientData.start)*Channels+c] = buffer1[w*c+x];
			}
			for( int x=0 ; x<_width ; x++ )
			{
				int xx = _width-1-x;
				for( int c=0 ; c<Channels ; c++) buffer1[xx*Channels+c] = buffer2[x*Channels+c];
			}
			for(int i=0;i<_cCount;i++)
			{
				int w = _clientSockets[i].clientData.end-_clientSockets[i].clientData.start;
				for( int x = 0 ; x < w ; x++ )
					for( int c = 0 ; c < Channels ; c++ )
						buffer2[w*c+x] = buffer1[(x+_clientSockets[i].clientData.start)*Channels+c];
				SendOnSocket( _clientSockets[i].client , buffer2 , sizeof(DataType)*w*Channels , "Failed to send restriction tail to client" );
			}
		}

		/////////////////////////////////////////
		// Now synchronize on the prolongation //
		/////////////////////////////////////////

		// Synchronize the head of the stream
		for(int k=0;k<_pPasses;k++)
		{
			for(int i=0;i<_cCount;i++)
			{
				int w = _clientSockets[i].clientData.end-_clientSockets[i].clientData.start;
				ReceiveOnSocket( _clientSockets[i].client , buffer1 , sizeof(DataType)*w*Channels , "Failed to get prolongation head from client" );
				for( int x = 0 ; x < w ; x++ )
					for( int c = 0; c < Channels ; c++ )
						buffer2[(x+_clientSockets[i].clientData.start)*Channels+c] = buffer1[w*c+x];
			}
			for(int x=0;x<_width/2;x++)
			{
				int xx = _width/2-1-x;
				for( int c=0 ; c<Channels ; c++ ) buffer1[xx*Channels+c] = buffer2[x*Channels+c];
			}
			for(int x=_width/2;x<_width;x++)
			{
				int xx = _width/2+_width-1-x;
				for( int c=0 ; c<Channels ; c++ ) buffer1[xx*Channels+c] = buffer2[x*Channels+c];
			}
			for(int i=0;i<_cCount;i++)
			{
				int w = _clientSockets[i].clientData.end-_clientSockets[i].clientData.start;
				for( int x = 0 ; x < w ; x++ )
					for( int c = 0 ; c < Channels ; c++ )
						buffer2[w*c+x] = buffer1[(x+_clientSockets[i].clientData.start)*Channels+c];
				SendOnSocket( _clientSockets[i].client , buffer2 , sizeof(DataType)*w*Channels , "Failed to send prolongation head to client" );
			}
		}
		// Synchronize the tail of the stream
		for(int k=0;k<_pPasses;k++)
		{
			for(int i=0;i<_cCount;i++)
			{
				int w = _clientSockets[i].clientData.end-_clientSockets[i].clientData.start;
				ReceiveOnSocket( _clientSockets[i].client , buffer1 , sizeof(DataType)*w*Channels , "Failed to get prolongation tail from client" );
				for( int x = 0 ; x < w ; x++ )
					for( int c = 0; c < Channels ; c++ )
						buffer2[(x+_clientSockets[i].clientData.start)*Channels+c] = buffer1[w*c+x];
			}
			for(int x=0;x<_width;x++)
			{
				int xx = _width-1-x;
				for( int c=0 ; c<Channels ; c++ ) buffer1[xx*Channels+c] = buffer2[x*Channels+c];
			}
			for(int i=0;i<_cCount;i++)
			{
				int w = _clientSockets[i].clientData.end-_clientSockets[i].clientData.start;
				for( int x = 0 ; x < w ; x++ )
					for( int c = 0 ; c < Channels ; c++ )
						buffer2[w*c+x] = buffer1[(x+_clientSockets[i].clientData.start)*Channels+c];
				SendOnSocket( _clientSockets[i].client , buffer2 , sizeof(DataType)*w*Channels , "Failed to send prolongation tail to client" );
			}
		}
	}
	FreeArray( buffer1 );
	FreeArray( buffer2 );
}
template<int Channels>
template<class DataType>
DWORD WINAPI SphericalSynchronizer<Channels>::RunThread(LPVOID lpParam)
{
	SphericalSynchronizer<Channels>* synchronizer = (SphericalSynchronizer*)lpParam;
	synchronizer->Run<DataType>();
	return 0;
}
/////////////////////////////
// SocketedMultigridServer //
/////////////////////////////
template< int Channels , class SyncType >
SocketedMultigridServer< Channels , SyncType >::SocketedMultigridServer(void)
{
	_synchronizerHandles = NULL;
	_icSynchronizers = NULL;
	_icSyncSockets = NULL;
}
template< int Channels , class SyncType >
SocketedMultigridServer< Channels , SyncType >::~SocketedMultigridServer(void)
{
	if(_globalData.spherical)
	{
		MyWaitForMultipleObjects( _oocLevels+1+_icLevels , _synchronizerHandles , 10000 , "SocketedMultigridServer< Channels , SyncType >::~SocketedMultigridServer" );
		for(int i=0;i<_oocLevels+1+_icLevels;i++)	CloseHandle( _synchronizerHandles[i] );
		delete[] _synchronizerHandles;
		_synchronizerHandles = NULL;
		if( _icSyncSockets )
		{
			for( int i=0;i<_icLevels ; i++ ) CloseSocket( _icSyncSockets[i] );
			delete[] _icSyncSockets;
		}
		_icSyncSockets = NULL;
	}
	if( _icSynchronizers )	delete[] _icSynchronizers;
	for( size_t i=0 ; i<_oocSynchronizers.size() ; i++ ) if( _oocSynchronizers[i] ) delete[] _oocSynchronizers[i];
	_oocSynchronizers.clear();
	_icSynchronizers = NULL;
}
template< int Channels , class SyncType >
bool SocketedMultigridServer< Channels , SyncType >::SetUp( char* address , int port , int clientCount , int iters , int inCoreRes , int minMGRes , int vCycles, int minBandSize ,
											   int tileWidth , int tileHeight , const char* tileExt , int quality , bool hdr , int lanes , bool verbose , bool spherical , bool sharpen , double iWeight , double gScale ,
											   bool blackOut , bool showProgress , int* clipDimensions )
{
	// Create a listening SOCKET for connecting to server
	SOCKET listenSocket = INVALID_SOCKET;
	listenSocket = GetListenSocket( port );
	if ( listenSocket == INVALID_SOCKET ) return false;
	if( address )
	{
		printf( "Server Address: %s:%d\n", address , port ) , fflush( stdout );
		return SetUp( address , listenSocket , clientCount , iters , inCoreRes , minMGRes , vCycles , minBandSize , tileWidth , tileHeight , tileExt , quality , hdr , lanes , verbose , spherical , sharpen , iWeight , gScale , blackOut , showProgress , clipDimensions );
	}
	else
	{
		char hostAddress[512];
		GetHostAddress( hostAddress );
		printf( "Server Address: %s:%d\n", hostAddress , port ) , fflush( stdout );
		return SetUp( hostAddress , listenSocket , clientCount , iters , inCoreRes , minMGRes , vCycles , minBandSize , tileWidth , tileHeight , tileExt , quality , hdr , lanes , verbose , spherical , sharpen , iWeight , gScale , blackOut , showProgress , clipDimensions );
	}
}
template< int Channels , class SyncType >
bool SocketedMultigridServer< Channels , SyncType >::SetUp( char* address , SOCKET listenSocket , int clientCount , int iters , int inCoreRes , int minMGRes , int vCycles , int minBandSize ,
											   int tileWidth , int tileHeight , const char* tileExt , int quality , bool hdr , int lanes , bool verbose , bool spherical , bool sharpen , double iWeight , double gScale ,
											   bool blackOut , bool showProgress , int* clipDimensions )
{
	ClientSocket* clientSockets = new ClientSocket[clientCount];

	_inCoreRes	= inCoreRes;
	_minMGRes	= minMGRes;

	_globalData.spherical	= spherical;
	_globalData.verbose		= verbose;
	_globalData.iters		= iters;
	_globalData.vCycles		= vCycles;
	_globalData.sharpen		= sharpen;
	_globalData.iWeight		= iWeight;
	_globalData.gScale		= gScale;
	_globalData.tileWidth	= tileWidth;
	_globalData.tileHeight	= tileHeight;
	_globalData.quality		= quality;
	_globalData.hdr			= hdr;
	_globalData.lanes		= lanes;
	_globalData.blackOut	= blackOut;
	_globalData.progress	= showProgress;
	if( tileExt ) strcpy( _globalData.tileExt , tileExt );
	else          _globalData.tileExt[0] = 0;


	// Establish a connection to each of the threads
	for( int i=0 ; i<clientCount ; i++ )
	{
		clientSockets[i].client  = AcceptSocket( listenSocket );
		printf( "Connected to thread: %d /%d     \r" , i+1 , clientCount );
	}
	printf( "\n" ) , fflush( stdout );

	// Get the client info
	int width = 0 , height;
	for( int i=0 ; i<clientCount ; i++ )
	{
		ReceiveOnSocket( clientSockets[i].client , GetPointer( clientSockets[i].clientData ) , sizeof(ClientData) , "Failed to get client data from client" ) ;
		if(!i) height = clientSockets[i].clientData.height;
		else if( height!=clientSockets[i].clientData.height ) fprintf( stderr , "Band heights differ: %d != %d\n" , height , clientSockets[i].clientData.height ) , exit(0);
		width += clientSockets[i].clientData.width;
	}

	// Sort the clients by index order
	qsort( clientSockets , clientCount , sizeof(ClientSocket) , ClientSocket::Sort );

	// Store the clipping width and then computed the padded dimensions
	if( clipDimensions )
	{
		_globalData.cWidth	= clipDimensions[0];
		_globalData.cHeight	= clipDimensions[1];
		_globalData.width	=  width;
		_globalData.height	= height;
	}
	else
	{
		_globalData.cWidth  = width;
		_globalData.cHeight = height;

		_globalData.width = width  , _globalData.height = height;

		SetPaddedSize( _globalData.width , _globalData.height , _minMGRes, _globalData.width , _globalData.height );
	}

	// Send the general solver information
	for( int i=0 ; i<clientCount ; i++ ) SendOnSocket( clientSockets[i].client , GetPointer( _globalData ) , sizeof(_globalData) , "Failed to send global data to client" );

	// Set the local information for the bands
	int start=0;
	for( int i=0 ; i<clientCount ; i++ )
	{
		int w = clientSockets[i].clientData.width;
		clientSockets[i].clientData.start = start;
		clientSockets[i].clientData.end = clientSockets[i].clientData.cEnd = start+w;
		if( i==clientCount-1 ) clientSockets[i].clientData.end  = _globalData.width;
		start = clientSockets[i].clientData.end;
		SendOnSocket( clientSockets[i].client , GetPointer( clientSockets[i].clientData ) , sizeof(ClientData) , "Failed to send client data to client" );
	}

	// Figure out how many different out-of-core levels are needed
	_oocLevels = 1;
	while( long long( width>>_oocLevels ) * long long( height>>_oocLevels ) >= _inCoreRes*_inCoreRes ) _oocLevels++;
	if( !ProcessPartition<int>::IsValidBandSize( width>>_oocLevels , height>>_oocLevels , iters , minBandSize ) )
	{
		fprintf( stderr , "In-core resolution is too small: %d x %d\n" , width>>_oocLevels , height>>_oocLevels );
		return false;
	}

	ProcessPartition< int > rootPartition;
	rootPartition.resize( clientCount );
	for( int i=0 ; i<clientCount ; i++ )
	{
		rootPartition[i].width = clientSockets[i].clientData.end - clientSockets[i].clientData.start;
		if( !ProcessPartition< int >::IsValidBandSize( rootPartition[i].width , _globalData.height , _globalData.iters , minBandSize ) )
		{
			fprintf( stderr , "Client[%d] has invalid dimension: %d x %d\n" , i , rootPartition[i].width , _globalData.height );
			return false;
		}
	}

	if( !processPartition.Initialize( rootPartition , _globalData.height , _globalData.iters , _oocLevels-1 , false , minBandSize ) )
	{
		fprintf( stderr , "Failed to create process partition\n" );
		return false;
	}
	for( int i=0 ; i<clientCount ; i++ )
	{
		int idx = i;
		int tCount = 1;
		for( int j=1 ; j<processPartition.size() && !(idx&1) && ( (idx>>1)<processPartition[j].size() ); j++ ) tCount++ , idx>>=1;
		SendOnSocket( clientSockets[i].client , GetPointer( tCount ) , sizeof(tCount) , "failed to send tCount to client" );
		Pointer( ProcessData ) pd = AllocArray< ProcessData >( tCount , 1 , "SocketedMultigridServer< Channels , SyncType >::SetUp (pd)" );
		pd[0].index = 0 ;
		pd[0].maxIndex = processPartition.size()-1;
		pd[0].offset = i;
		pd[0].maxOffset = clientCount-1;
		pd[0].startDepth = processPartition[0].startDepth;
		pd[0].endDepth = processPartition[0].endDepth;
		processPartition[0].setBounds( i , pd[0].start , pd[0].stop );
		pd[0].width = _globalData.width;
		pd[0].height = _globalData.height;
		pd[0].children = processPartition[0][i].children.size();

		idx = i;
		for( int j=1 ; j<tCount ; j++ )
		{
			idx >>= 1;
			pd[j].index  = j;
			pd[j].maxIndex = processPartition.size()-1;
			pd[j].offset = idx;
			pd[j].maxOffset = processPartition[j].size()-1;
			pd[j].startDepth = processPartition[j].startDepth;
			pd[j].endDepth = processPartition[j].endDepth;
			processPartition[j].setBounds( idx , pd[j].start , pd[j].stop );
			pd[j].width = _globalData.width >> processPartition[j].startDepth;
			pd[j].height = _globalData.height >> processPartition[j].startDepth;
			pd[j].children = processPartition[j][idx].children.size();
		}
		SendOnSocket( clientSockets[i].client , pd , sizeof(ProcessData) * tCount , "Failed to send process data to client" );
		FreeArray( pd );
	}
	delete[] clientSockets;

	// At this point we transition our communications so that we are talking to the threads, not processes
	int totalPartitions = 0;
	for( int i=0 ; i<processPartition.size() ; i++ ) totalPartitions += processPartition[i].size();
	int partitionIndex = 0;
	for( int i=0 ; i<processPartition.size() ; i++ )
		for( int j=0 ; j<processPartition[i].size() ; j++ )
		{
			// Establish a connection to the process
			int index , offset;
			SOCKET sock = AcceptSocket( listenSocket );
			partitionIndex++;
			printf( "Connected to block: %d /%d     \r" , partitionIndex , totalPartitions );
			ReceiveOnSocket( sock , GetPointer( index  ) , sizeof(index) , "Failed to get client index" );
			ReceiveOnSocket( sock , GetPointer( offset ) , sizeof(offset) , "Failed to get client offset" );
			processPartition[index][offset].data.sock = sock;
			if( index )
			{
				ReceiveOnSocket ( sock , GetPointer( processPartition[index][offset].data.childAddressX ) , sizeof(IPData) , "Failed to get child address X" );
				ReceiveOnSocket ( sock , GetPointer( processPartition[index][offset].data.childAddressB ) , sizeof(IPData) , "Failed to get child address B" );
			}
			if( spherical || offset )
			{
				ReceiveOnSocket ( sock , GetPointer( processPartition[index][offset].data.leftAddress ) , sizeof(IPData) , "Failed to get left address from client" );
			}
		}
	printf( "\n" ) , fflush( stdout );

	// Create a listening socket to establish the connection between the in-core and out-of-core solver
	int listenerCount = 2;
	HANDLE listenerHandles[3];

	_inCoreConnectInfo.init( processPartition[processPartition.size()-1].size() );

	IPData inCoreAddressX , inCoreAddressB , inCoreAddressLeft;

	{
		MyWinSock::SystemLock lock;
		memcpy( &inCoreAddressX.address    , gethostbyname(address)->h_addr , sizeof(struct in_addr) );
		memcpy( &inCoreAddressB.address    , gethostbyname(address)->h_addr , sizeof(struct in_addr) );
		memcpy( &inCoreAddressLeft.address , gethostbyname(address)->h_addr , sizeof(struct in_addr) );
	}
	_inCoreConnectInfo.listenChildX = GetListenSocket( inCoreAddressX.port );
	_inCoreConnectInfo.listenChildB = GetListenSocket( inCoreAddressB.port );
	if ( _inCoreConnectInfo.listenChildX == INVALID_SOCKET ) return false;
	if ( _inCoreConnectInfo.listenChildB == INVALID_SOCKET ) return false;
	listenerHandles[0] = SpawnListenerSocket( _inCoreConnectInfo.listenChildX , _inCoreConnectInfo.childX , processPartition[processPartition.size()-1].size() );
	listenerHandles[1] = SpawnListenerSocket( _inCoreConnectInfo.listenChildB , _inCoreConnectInfo.childB , processPartition[processPartition.size()-1].size() );
	if( !listenerHandles[0] ) return false;
	if( !listenerHandles[1] ) return false;
	SOCKET leftSocket = INVALID_SOCKET;
	if( spherical )
	{
		_inCoreConnectInfo.listenLeft = GetListenSocket( inCoreAddressLeft.port );
		if ( _inCoreConnectInfo.listenLeft == INVALID_SOCKET ) return false;
		listenerCount++;
		listenerHandles[2] = SpawnListenerSocket( _inCoreConnectInfo.listenLeft , &leftSocket , 1 );
		if( !listenerHandles[2] ) return false;
	}

	// Now send everybody information about who to connect to for their data
	for( int i=0 ; i<processPartition.size() ; i++ )
		for( int j=0 ; j<processPartition[i].size() ; j++ )
		{
			if( i )
				for( size_t k=0 ; k<processPartition[i][j].children.size() ; k++ )
				{
					SOCKET& sock = processPartition[i-1][processPartition[i][j].children[k]].data.sock;
					SendOnSocket ( sock , GetPointer( processPartition[i][j].data.childAddressX ) , sizeof(IPData) , "Failed to send child address X to client" );
					SendOnSocket ( sock , GetPointer( processPartition[i][j].data.childAddressB ) , sizeof(IPData) , "Failed to send child address B to client" );
			}
		}
	{
		int i = processPartition.size()-1;
		for( int j=0; j<processPartition[i].size() ; j++ )
		{
			SOCKET& sock = processPartition[i][j].data.sock;
			SendOnSocket ( sock , GetPointer( inCoreAddressX ) , sizeof(IPData) , "Failed to send address X to client" );
			SendOnSocket ( sock , GetPointer( inCoreAddressB ) , sizeof(IPData) , "Failed to send address B to client" );
		}
	}
	for( int i=0 ; i<processPartition.size() ; i++ )
		for( int j=0 ; j<processPartition[i].size() ; j++ )
			if( spherical || j )
			{
				SOCKET& sock = processPartition[i][(j-1+processPartition[i].size())%processPartition[i].size()].data.sock;
				SendOnSocket ( sock , GetPointer( processPartition[i][j].data.leftAddress ) , sizeof(IPData), "Failed tos send left address to client" );
			}
	if( spherical )
	{
		_inCoreConnectInfo.right = new Socket( inCoreAddressLeft.address , inCoreAddressLeft.port , 5 , false );
	}
	MyWaitForMultipleObjects( listenerCount , listenerHandles , 10000 , "SocketedMultigridServer< Channels , SyncType >::SetUp" );
	if( spherical ) _inCoreConnectInfo.left = new Socket( leftSocket );

	// Figure out how many different in-core resolutions are needed
	_icLevels = 1;
	int w = _globalData.width >> _oocLevels , h = _globalData.height >> _oocLevels;
	while( ProcessPartition<int>::IsValidBandSize( w>>_icLevels , h>>_icLevels , _globalData.iters , 0 ) && (w>>_icLevels)>=_minMGRes && (h>>_icLevels)>=_minMGRes )
		_icLevels++;

	// Now for the spherical synchronizers
	if( _globalData.spherical )
	{
		_icSynchronizers = new SphericalSynchronizer< Channels > [ _icLevels ];
		if( !_icSynchronizers )
		{
			fprintf(stderr,"Failed to allocate in-core synchronizers\n");
			return false;
		}
		_oocSynchronizers.resize( processPartition.size() );
		for( int i=0 ; i<processPartition.size() ; i++ )
		{
			if( !i ) _oocSynchronizers[i] = new SphericalSynchronizer< Channels >[ (processPartition[i].endDepth-processPartition[i].startDepth+1)+1 ];
			else	 _oocSynchronizers[i] = new SphericalSynchronizer< Channels >[ (processPartition[i].endDepth-processPartition[i].startDepth+1)   ];
		}

		int *oocWidths , icWidths;
		oocWidths = new int[clientCount];

		for( int i=0 ; i<processPartition.size() ; i++ )
		{
			if(!i)
			{
				for( int j=0 ; j<processPartition[i].size() ; j++ ) oocWidths[j] = processPartition[i][j].width;
				imageSynchronizer.init( _globalData.width , oocWidths , clientCount , Degree , 0 , 1 );
			}
			for( int d=processPartition[i].startDepth ; d<=processPartition[i].endDepth ; d++ )
			{
				int depth = d-processPartition[i].startDepth;
				for( int j=0 ; j<processPartition[i].size() ; j++ ) oocWidths[j] = processPartition[i][j].width >> depth;
				if( _globalData.verbose )
					_oocSynchronizers[i][depth].init( _globalData.width>>d , oocWidths , processPartition[i].size() , (_globalData.iters*Degree+Degree)+Degree , (_globalData.iters*Degree+Degree)+Degree , _globalData.vCycles );
				else
					_oocSynchronizers[i][depth].init( _globalData.width>>d , oocWidths , processPartition[i].size() , (_globalData.iters*Degree+Degree)+Degree , (_globalData.iters*Degree+Degree)+     0 , _globalData.vCycles );
			}
		}
		delete[] oocWidths;
		for( int d=0 ; d<_icLevels ; d++ )
		{
			int depth = d+_oocLevels;
			icWidths = _globalData.width >> depth;
			if( _globalData.verbose )
				_icSynchronizers[d].init( icWidths , &icWidths , 1 , (_globalData.iters*Degree+Degree)+Degree , (_globalData.iters*Degree+Degree)+Degree , _globalData.vCycles );
			else
				_icSynchronizers[d].init( icWidths , &icWidths , 1 , (_globalData.iters*Degree+Degree)+Degree , (_globalData.iters*Degree+Degree)+     0 , _globalData.vCycles );
		}
		_synchronizerHandles = new HANDLE[ _oocLevels + 1 + _icLevels ];

		int idx = 0;
		{
			DWORD synchThreadID;
			if( _globalData.sharpen )
			{
				_synchronizerHandles[idx]=CreateThread(
					NULL,													// default security attributes
					0,														// use default stack size  
					SphericalSynchronizer<Channels>::RunThread< SyncType >,	// thread function 
					&imageSynchronizer,									 	// argument to thread function 
					0,														// use default creation flags 
					&synchThreadID);										// returns the thread identifier
			}
			else
			{
				_synchronizerHandles[idx]=CreateThread(
					NULL,																					// default security attributes
					0,																						// use default stack size  
					SphericalSynchronizer<Channels>::RunThread< ImageData< SyncType , unsigned __int16 > >,	// thread function 
					&imageSynchronizer,																		// argument to thread function 
					0,																						// use default creation flags 
					&synchThreadID);																		// returns the thread identifier
			}
			if(!_synchronizerHandles[idx++]) fprintf(stderr,"Failed to create sync thread\n") , exit(0);
		}
		for( int i=0 ; i<processPartition.size() ; i++ )
		{
			int depths = processPartition[i].endDepth - processPartition[i].startDepth + 1;
			for( int d=0 ; d<depths ; d++ )
			{
				DWORD synchThreadID;
				_synchronizerHandles[idx]=CreateThread( 
					NULL,													// default security attributes
					0,														// use default stack size  
					SphericalSynchronizer<Channels>::RunThread< SyncType >,	// thread function 
					&_oocSynchronizers[i][d],								// argument to thread function 
					0,														// use default creation flags 
					&synchThreadID);										// returns the thread identifier
				if( !_synchronizerHandles[idx++] ) fprintf(stderr,"Failed to create sync thread\n") , exit(0);
			}
		}
		for(int i=0;i<_icLevels;i++)
		{
			DWORD synchThreadID;
			_synchronizerHandles[idx]=CreateThread( 
				NULL,													// default security attributes
				0,														// use default stack size  
				SphericalSynchronizer<Channels>::RunThread< float >,	// thread function 
				&_icSynchronizers[i],									// argument to thread function 
				0,														// use default creation flags 
				&synchThreadID);										// returns the thread identifier
			if( !_synchronizerHandles[idx++] ) fprintf(stderr,"Failed to create sync thread\n") , exit(0);
		}

		int* icPorts = new int[_icLevels];
		for( int d=0 ; d<_icLevels ; d++ )	icPorts[d] = _icSynchronizers[d].port(0);

		for( int i=0 ; i<processPartition.size() ; i++ )
		{
			int depths = processPartition[i].endDepth - processPartition[i].startDepth + 1;
			Pointer( int ) oocPorts = AllocArray< int >( depths , 1 , "SocketedMultigridServer< Channels , SyncType >::SetUp (oocPorts)" );
			for( int j=0 ; j< processPartition[i].size() ; j++ )
			{
				if( !i )
				{
					int imagePort = imageSynchronizer.port(j);
					SendOnSocket( processPartition[i][j].data.sock , GetPointer( imagePort ) , sizeof(int) , "Failed to send image port to client" );
				}
				for( int d=0 ; d<depths ; d++ ) oocPorts[d] = _oocSynchronizers[i][d].port(j);
				SendOnSocket( processPartition[i][j].data.sock , oocPorts , sizeof(int)*depths , "Failed to send out of core ports to client" );
			}
			FreeArray( oocPorts );
		}
		_icSyncSockets = new SOCKET[_icLevels];
		for( int i=0 ; i<_icLevels ; i++ )
		{
			_icSyncSockets[i] = GetConnectSocket( address , icPorts[i] );
			if( _icSyncSockets[i] == INVALID_SOCKET )	return false;
		}
		delete[] icPorts;
	}

	CloseSocket( listenSocket );
	return true;
}
template< int Channels , class SyncType >
void SocketedMultigridServer< Channels , SyncType >::Run(void)
{
#if USE_DENORMAL_CONTROL
//	unsigned int control_word_x87  , control_word_sse2;
//	__control87_2( _DN_FLUSH , _MCW_DN , &control_word_x87 , &control_word_sse2 );
	_controlfp( _DN_FLUSH , _MCW_DN );
#endif // USE_DENORMAL_CONTROL
	StreamingGrid *inB , *outX;
	printf( "Running server\n" ) , fflush( stdout );
	DotProductStencil stencils[4];
	Pointer( SolverInfo< Channels > )  solverInfo = AllocArray< SolverInfo< Channels > >( _oocLevels , 1 , "SocketedMultigridServer< Channels , SyncType >::Run (solverInfo)" );
	Pointer( SolverInfo< Channels > ) tSolverInfo = AllocArray< SolverInfo< Channels > >( _oocLevels , 1 , "SocketedMultigridServer< Channels , SyncType >::Run (tSolverInfo)" );
	double zeroAverage[Channels];
	for(int c=0;c<Channels;c++)	zeroAverage[c]=0;
	Vector<float> globalB,globalX;
	int gWidth  = _globalData.width  >> _oocLevels;
	int gHeight = _globalData.height >> _oocLevels;

	globalB.Resize( gWidth * gHeight * Channels );
	globalX.Resize( gWidth * gHeight * Channels );

	memset(_average, 0 , sizeof(_average) );
	inB  = GetMultiSocketBackedGrid< Channels , SyncType > ( _inCoreConnectInfo.childB , _inCoreConnectInfo.children , gHeight    , false , true );
	outX = GetMultiSocketBackedGrid< Channels , SyncType > ( _inCoreConnectInfo.childX , _inCoreConnectInfo.children , gHeight<<1 , false , true );

	for(int ii=0;ii<_globalData.vCycles;ii++)
	{
		double t;

		t=Time();
		memset( solverInfo , 0 , sizeof( SolverInfo<Channels> ) * _oocLevels );
		inB->reset( true , 1 );

		for( int j=0 ; j<gHeight ; j++ )
		{
			Pointer( SyncType ) row = ( Pointer( SyncType ) )(*inB)[j];
			for( int i=0 ; i<Channels*gWidth ; i++ ) globalB[gWidth*j*Channels + i ] = float( row[i] );
			inB->advance();
		}
		for( int i=0 ; i<processPartition.size() ; i++ )
			for( int j=0 ; j<processPartition[i].size() ; j++ )
			{
				ReceiveOnSocket( processPartition[i][j].data.sock , tSolverInfo + processPartition[i].startDepth , sizeof(SolverInfo<Channels>) * (processPartition[i].endDepth-processPartition[i].startDepth+1) , "Failed to get process data from process" );
				for( int d=processPartition[i].startDepth ; d<=processPartition[i].endDepth ; d++ )
				{
					solverInfo[d].bSquareNorm += tSolverInfo[d].bSquareNorm;
					solverInfo[d].rSquareNorm += tSolverInfo[d].rSquareNorm;
					solverInfo[d].xSquareNorm += tSolverInfo[d].xSquareNorm;
					for( int c=0 ; c<Channels ; c++ ) solverInfo[d].solutionSum[c] += tSolverInfo[d].solutionSum[c];
				}
			}
		for( int j=0 ; j<processPartition[0].size() ; j++ )
		{
			double avg[Channels];
			ReceiveOnSocket ( processPartition[0][j].data.sock , GetPointer( avg ) , sizeof(avg) , "Failed to get average from process" );
			for( int c=0 ; c<Channels ; c++ )	_average[c]+=avg[c];
		}
		if(ii==0)	for( int c=0 ; c<Channels ; c++ )	_average[c] /= _globalData.cWidth	,	_average[c] /= _globalData.cHeight;
		printf("Client Restriction:    %f\n",Time()-t) , fflush( stdout );
		if(_globalData.verbose)
			for( int i=0 ; i<processPartition.size() ; i++ )
				for( int d=processPartition[i].startDepth ; d<=processPartition[i].endDepth ; d++ )
				{
					int dd = processPartition[i].endDepth - ( d - processPartition[i].startDepth );
					printf("\tError[%d x %d] %g -> %g\t(%g)\n" , _globalData.width >> d , _globalData.height >> d , sqrt(solverInfo[dd  ].bSquareNorm) , sqrt(solverInfo[dd  ].rSquareNorm) , sqrt(solverInfo[dd  ].xSquareNorm) ) , fflush( stdout );
				}

		// Get the base solution
		DotProductStencil dotMajor , d2DotMajor , dotMinor ,d2DotMinor;
		SetDownSampledStencil( _globalData.width  , _oocLevels , dotMajor , d2DotMajor );
		SetDownSampledStencil( _globalData.height , _oocLevels , dotMinor , d2DotMinor );
		if( ii==_globalData.vCycles-1)	SolveInCore( dotMajor , d2DotMajor , dotMinor ,d2DotMinor , globalB , globalX , _average);
		else							SolveInCore( dotMajor , d2DotMajor , dotMinor ,d2DotMinor , globalB , globalX , zeroAverage);

		t=Time();
		int gWidth2 = gWidth<<1 , gHeight2 = gHeight<<1;
		outX->reset( false , 1 );
		// Get the up-sampled solution
		float *upX = new float[ Channels * gWidth2 * gHeight2 ];

		for( int c=0 ; c<Channels ; c++ )
		{
			Vector< float > low , high;
			low.Resize( gWidth * gHeight );
			for( int j=0 ; j<gHeight  ; j++ ) memcpy( &low[ j*gWidth ] , &globalX[ j*gWidth*Channels + gWidth*c ] ,  sizeof(float) * gWidth );
			if( _globalData.spherical ) UpSampleSphericalSolution  ( low , gWidth , gHeight , high );
			else						UpSampleRectangularSolution( low , gWidth , gHeight , high );
			for( int j=0 ; j<gHeight2 ; j++ ) memcpy( &upX[ j*gWidth2*Channels + c*gWidth2 ] , &high[ j*gWidth2 ] ,  sizeof(float) * gWidth2 );
		}

		// Send the up-sample solution to the socket
		for( int j=0 ; j<gHeight2 ; j++ )
		{
			Pointer( SyncType ) row = ( Pointer( SyncType ) )(*outX)[j];
			for( int i=0 ; i<Channels*gWidth2 ; i++ ) row[i] = SyncType( upX[gWidth2*j*Channels+i] );
			outX->advance();
		}
		delete upX;

		memset( solverInfo , 0 , sizeof(SolverInfo< Channels >) * _oocLevels );
		for( int i=0 ; i<processPartition.size() ; i++ )
			for( int j=0 ; j<processPartition[i].size() ; j++ )
			{
				ReceiveOnSocket ( processPartition[i][j].data.sock , tSolverInfo + processPartition[i].startDepth , sizeof(SolverInfo<Channels>) * (processPartition[i].endDepth-processPartition[i].startDepth+1) , "Failed to get process data from process" );
				for( int d=processPartition[i].startDepth ; d<=processPartition[i].endDepth ; d++ )
				{
					solverInfo[d].bSquareNorm += tSolverInfo[d].bSquareNorm;
					solverInfo[d].rSquareNorm += tSolverInfo[d].rSquareNorm;
					solverInfo[d].xSquareNorm += tSolverInfo[d].xSquareNorm;
					for( int c=0 ; c<Channels ; c++ ) solverInfo[d].solutionSum[c] += tSolverInfo[d].solutionSum[c];
				}
			}
		printf("Client Prolongation:    %f\n",Time()-t) , fflush( stdout );
		if( _globalData.verbose )
			for( int i=processPartition.size()-1 ; i>=0 ; i-- )
				for( int d=processPartition[i].endDepth ; d>=processPartition[i].startDepth ; d-- )
				{
					int dd = processPartition[i].startDepth - ( d - processPartition[i].endDepth );
					printf("\tError[%d x %d] %g -> %g\n" , _globalData.width >> d , _globalData.height >> d , sqrt(solverInfo[dd  ].bSquareNorm) , sqrt(solverInfo[dd  ].rSquareNorm)) , fflush( stdout );
				}
	}
	FreeArray( solverInfo  );
	FreeArray( tSolverInfo );
	delete inB;
	delete outX;

}
template< int Channels , class SyncType >
void SocketedMultigridServer< Channels , SyncType >::SolveInCore(DotProductStencil& dotMajor,DotProductStencil& d2DotMajor,
																 DotProductStencil& dotMinor,DotProductStencil& d2DotMinor,
																 Vector<float>& in,Vector<float>& out,double average[Channels])
{
	double t;
	int w = _globalData.width >> _oocLevels , h = _globalData.height >> _oocLevels;
	SocketedMultiGridStreamingSolver< Channels , float , float >* solvers;
	DotProductStencil lowDotMajor , lowD2DotMajor , lowDotMinor , lowD2DotMinor;

	SetDownSampledStencil( dotMajor , d2DotMajor , _globalData.width  >> _oocLevels , _icLevels-1 , lowDotMajor , lowD2DotMajor );
	SetDownSampledStencil( dotMinor , d2DotMinor , _globalData.height >> _oocLevels , _icLevels-1 , lowDotMinor , lowD2DotMinor );
	Vector<float> lowB,lowX;
	StreamingGrid *B,*X;

	solvers=new SocketedMultiGridStreamingSolver< Channels , float , float >[_icLevels];

	for( int i=1 ; i<_icLevels   ; i++ )	solvers[i].parent=&solvers[i-1];
	for( int i=0 ; i<_icLevels-1 ; i++ )	solvers[i].rChild =solvers[i].pChild =&solvers[i+1];
	for( int i=0 ; i<_icLevels   ; i++ )
	{
		solvers[i].bSquareNorm=solvers[i].rSquareNorm=solvers[i].xSquareNorm=0;
		for( int c=0 ; c<Channels ; c++ ) solvers[i].solutionSum[c]=0;
		solvers[i].setResidual = true;
	}
	if( _globalData.sharpen )	solvers[_icLevels-1].Initialize( dotMajor , d2DotMajor , dotMinor , d2DotMinor , _globalData.iWeight , 0 , w , w , h , _globalData.iters , _inCoreConnectInfo.left , _icSyncSockets , _inCoreConnectInfo.right , false , _globalData.spherical , NULL );
	else						solvers[_icLevels-1].Initialize( dotMajor , d2DotMajor , dotMinor , d2DotMinor ,                       0 , w , w , h , _globalData.iters , _inCoreConnectInfo.left , _icSyncSockets , _inCoreConnectInfo.right , false , _globalData.spherical , NULL );

	lowX.Resize(solvers[0].major*solvers[0].minor*Channels);
	lowB.Resize(solvers[0].major*solvers[0].minor*Channels);

#if ASSERT_ARRAY_ACCESS
	B = new MemoryBackedGrid( Array< float >::FromPointer( &lowB[0] , lowB.Dimensions() ) , solvers[0].major*Channels*sizeof(float) , solvers[0].minor , true );
	X = new MemoryBackedGrid( Array< float >::FromPointer( &lowX[0] , lowX.Dimensions() ) , solvers[0].major*Channels*sizeof(float) , solvers[0].minor , true );
	solvers[_icLevels-1].inB = new MemoryBackedGrid( Array< float >::FromPointer(  &in[0] ,  in.Dimensions() ) , w*Channels*sizeof(float) , h , true );
	solvers[_icLevels-1].inX = new MemoryBackedGrid( Array< float >::FromPointer( &out[0] , out.Dimensions() ) , w*Channels*sizeof(float) , h , true );
#else // !ASSERT_ARRAY_ACCESS
	B = new MemoryBackedGrid( ( Pointer( byte ) )&lowB[0] , solvers[0].major * Channels * sizeof(float) , solvers[0].minor , true );
	X = new MemoryBackedGrid( ( Pointer( byte ) )&lowX[0] , solvers[0].major * Channels * sizeof(float) , solvers[0].minor , true );
	solvers[_icLevels-1].inB = new MemoryBackedGrid( (Pointer( byte ) ) &in[0] , w*Channels*sizeof(float) , h , true );
	solvers[_icLevels-1].inX = new MemoryBackedGrid( (Pointer( byte ) )&out[0] , w*Channels*sizeof(float) , h , true );
#endif // ASSERT_ARRAY_ACCESS
	solvers[0].outX = X;
	solvers[0].outB = B;

	// Data for the interleaved streaming multigrid
	t=Time();
	// Initialize
	solvers[_icLevels-1].InitRestriction();
	solvers[_icLevels-1].SetRestriction();
	solvers[_icLevels-1].SolveRestriction();

	// Clean up
	delete solvers[_icLevels-1].inB;
	solvers[_icLevels-1].inB=NULL;
	delete solvers[_icLevels-1].inX;
	solvers[_icLevels-1].inX=NULL;

	printf( "Server Restriction:        %f\n" , Time()-t ) , fflush( stdout );
	if(_globalData.verbose)
		for( int i=_icLevels-1 ; i>=0 ; i-- )
			printf("\tError[%d x %d] %g -> %g\n",solvers[i].major,solvers[i].minor,sqrt(solvers[i].bSquareNorm),sqrt(solvers[i].rSquareNorm)) , fflush( stdout );

	solvers[_icLevels-1].UnSetRestriction();
	// Get the base solution
	SparseMatrix<double> lap;
	t=Time();
	if( _globalData.spherical )
		if(!_globalData.sharpen || _globalData.iWeight==0)	SetSphericalLaplacianMatrix(lowDotMajor,lowD2DotMajor,lowDotMinor,lowD2DotMinor,lap,solvers[0].major,solvers[0].minor);
		else												SetSphericalLaplacianMatrix(lowDotMajor,lowD2DotMajor,lowDotMinor,lowD2DotMinor,lap,solvers[0].major,solvers[0].minor,_globalData.iWeight);
	else
		if(!_globalData.sharpen || _globalData.iWeight==0)	SetRectangularLaplacianMatrix(lowDotMajor,lowD2DotMajor,lowDotMinor,lowD2DotMinor,lap,solvers[0].major,solvers[0].minor);
		else												SetRectangularLaplacianMatrix(lowDotMajor,lowD2DotMajor,lowDotMinor,lowD2DotMinor,lap,solvers[0].major,solvers[0].minor,_globalData.iWeight);
	{
		Vector<double> myLowX,myLowB;
		myLowB.Resize(solvers[0].major*solvers[0].minor);
		myLowX.Resize(solvers[0].major*solvers[0].minor);
		for(int c=0;c<Channels;c++)
		{
			for(int i=0;i<solvers[0].major;i++)
				for(int j=0;j<solvers[0].minor;j++)
				{
					myLowX[i+j*solvers[0].major] = lowX[i+j*solvers[0].major*Channels+c*solvers[0].major];
					myLowB[i+j*solvers[0].major] = lowB[i+j*solvers[0].major*Channels+c*solvers[0].major];
				}
			myLowB = myLowB - lap * myLowX;
			myLowX.SetZero();

		if(!_globalData.sharpen || _globalData.iWeight==0)	MySolveConjugateGradient<Type>(lap,myLowB,6*int(sqrt(lap.groups+1.0)),myLowX);
		else												MySolveConjugateGradient<Type>(lap,myLowB,6*int(sqrt(lap.groups+1.0)),myLowX);

		for(int i=0;i<solvers[0].major;i++)
			for(int j=0;j<solvers[0].minor;j++)
			{
				lowX[i+j*solvers[0].major*Channels+c*solvers[0].major] = float( myLowX[i+j*solvers[0].major] );
				myLowX[i+j*solvers[0].major] = 0.0;
			}
		}
	}
	printf( "Server Conjugate Gradient: %f\n" , Time()-t ) , fflush( stdout );

	for(int i=0;i<_icLevels;i++)
	{
		solvers[i].bSquareNorm=solvers[i].rSquareNorm=solvers[i].xSquareNorm=0;
		for(int c=0;c<Channels;c++)	solvers[i].solutionSum[c]=0;
		solvers[i].setResidual=_globalData.verbose;
	}

	// Solve the prolongation
	t=Time();
	solvers[0].inX=X;
	// Set the child dependencies
#if ASSERT_ARRAY_ACCESS
	solvers[_icLevels-1].outX = new MemoryBackedGrid( Array< float >::FromPointer( &out[0] , out.Dimensions() ) , w*Channels*sizeof(float) , h , true );
#else // !ASSERT_ARRAY_ACCESS
	solvers[_icLevels-1].outX = new MemoryBackedGrid( (Pointer( byte ) )&out[0],w*Channels*sizeof(float),h , true );
#endif // ASSERT_ARRAY_ACCESS
	solvers[_icLevels-1].InitProlongation();
	solvers[0].SetProlongation();
	// Solve
	solvers[0].SolveProlongation();
	printf( "Server Prolongation:       %f\n" , Time()-t ) , fflush( stdout );
	if(_globalData.verbose)
	{
		for(int i=0;i<_icLevels;i++)
		{
			printf("\tError[%d x %d] %g -> %g\n",solvers[i].major,solvers[i].minor,sqrt(solvers[i].bSquareNorm),sqrt(solvers[i].rSquareNorm)) , fflush( stdout );
		}
	}
	solvers[0].UnSetProlongation();
	delete solvers[_icLevels-1].outX;
	delete[] solvers;
	delete B;
	delete X;
	B=X=NULL;

	int ww=(long long(_globalData.cWidth) *w+_globalData.width -1)/_globalData.width;
	int hh=(long long(_globalData.cHeight)*h+_globalData.height-1)/_globalData.height;
	for(int c=0;c<Channels;c++)
	{
		double newAverage=0;
		int pCount=0;

		for(int j=0;j<hh;j++)
		{
			float* o=&out[j*w*Channels+c*w];
			for(int i=0;i<ww;i++)
			{
				newAverage+=o[i];
				pCount++;
			}
		}

		newAverage/=pCount;
		newAverage=average[c]-newAverage;

		if(!_globalData.sharpen || _globalData.iWeight==0)
			for(int j=0;j<h;j++)
			{
				float* o=&out[j*w*Channels+c*w];
				for(int i=0;i<w;i++)	o[i]+=newAverage;
			}
	}
}

//////////////////////////////////
// SocketedSuperMultigridServer //
//////////////////////////////////
template<int Channels>
bool SocketedSuperMultigridServer<Channels>::SetUp( char* prefix , int port , int clientCount , int iters , int inCoreRes , int minMGRes , int vCycles , int minBandSize ,
												    int tileWidth , int tileHeight , const char* tileExt , int quality , bool hdr , int lanes , bool verbose , bool spherical , bool sharpen , double iWeight , double gScale ,
													bool blackOut , bool showProgress )
{
	_port = port;
	_clientCount = clientCount;
	_iters = iters;
	_inCoreRes = inCoreRes;
	_minMGRes = minMGRes;
	_vCycles = vCycles;
	_tileWidth = tileWidth;
	_tileHeight = tileHeight;
	if( tileExt ) strcpy( _tileExt , tileExt );
	else          _tileExt[0] = 0;
	_quality = quality;
	_hdr = hdr;
	_lanes = lanes;
	_verbose = verbose;
	_spherical = spherical;
	_sharpen = sharpen;
	_iWeight = iWeight;
	_gScale = gScale;
	_blackOut = blackOut;
	_showProgress = showProgress;
	_minBandSize = minBandSize;
	if( _minBandSize > _inCoreRes ) _inCoreRes = _minBandSize;

	GetHostAddress( _address , prefix );

	return true;
}
template<int Channels>
bool SocketedSuperMultigridServer<Channels>::Run( void )
{
	return _Run< float >( );
}
template< int Channels >
template< class SyncType >
bool SocketedSuperMultigridServer< Channels >::_Run( void )
{
	GlobalData globalData;
	int subClientCount = 0;

	globalData.spherical	= _spherical;
	globalData.verbose		= _verbose;
	globalData.iters		= _iters;
	globalData.vCycles		= _vCycles;
	globalData.sharpen		= _sharpen;
	globalData.iWeight		= _iWeight;
	globalData.tileWidth	= _tileWidth;
	globalData.tileHeight	= _tileHeight;
	globalData.gScale		= _gScale;
	globalData.quality		= _quality;
	globalData.hdr			= _hdr;
	globalData.lanes		= _lanes;
	globalData.blackOut		= _blackOut;
	globalData.progress		= _showProgress;
	strcpy( globalData.tileExt , _tileExt );

	// Create a listening SOCKET for connecting to server
	SOCKET listenSocket = INVALID_SOCKET;
	listenSocket = GetListenSocket( _port );
	if ( listenSocket == INVALID_SOCKET ) return false;
	printf( "Valid Addresses\n" ) , fflush( stdout );
	PrintHostAddress();
	printf( "Server Address: %s:%d\n", _address , _port ) , fflush( stdout );

	ClientSocket* clientSockets = new ClientSocket[_clientCount];
	int coreCount = 0;

	// Establish a connection to the clients
	for( int i=0 ; i<_clientCount ; i++ )
	{
		clientSockets[i].client  = AcceptSocket( listenSocket );
		printf( "Connected to process: %d /%d     \r" , i+1 , _clientCount );
	}
	printf( "\n" ) , fflush( stdout );
	// Get the client info
	int width = 0 , height;
	for( int i=0 ; i<_clientCount ; i++ )
	{
		ReceiveOnSocket( clientSockets[i].client , GetPointer( clientSockets[i].clientData ) , sizeof(ClientData) , "Failed to get client data from client" );
		if(!i) height = clientSockets[i].clientData.height;
		else if( height!=clientSockets[i].clientData.height ) fprintf( stderr , "Band heights differ: %d != %d\n" , height , clientSockets[i].clientData.height ) , exit(0);
		width += clientSockets[i].clientData.width;
		subClientCount += clientSockets[i].clientData.subClients;
	}
	// Sort the clients by index order
	qsort( clientSockets , _clientCount , sizeof(ClientSocket) , ClientSocket::Sort );

	// Store the clipping width and then computed the padded dimensions
	globalData.cWidth  = width;
	globalData.cHeight = height;


	globalData.width = width  , globalData.height = height;

	SetPaddedSize( globalData.width , globalData.height , _minMGRes, globalData.width , globalData.height );
	int oocLevels = 1;

	while( long long( globalData.width >> oocLevels ) * long long( globalData.height >> oocLevels ) >= _inCoreRes*_inCoreRes ) oocLevels++;
	if( !ProcessPartition<int>::IsValidBandSize( globalData.width>>oocLevels , globalData.height>>oocLevels , globalData.iters , _minBandSize ) )
	{
		fprintf( stderr , "In-core resolution is too small: %d x %d\n" , globalData.width>>oocLevels , globalData.height>>oocLevels );
		return false;
	}
	int blockSize = 4<<oocLevels;
	blockSize >>= 1;

	for( int i=0 ; i<_clientCount-1 ; i++ )
		if( clientSockets[i].clientData.width % blockSize )
		{
			fprintf( stderr , "Non-paddable client[%d] width not multiple of block size: %d mod %d != 0 \n" , i , clientSockets[i].clientData.width , blockSize );
			exit(0);
		}

	printf( "  Image Dimension: %d x %d -> %d x %d\n" , width , height , globalData.width , globalData.height ) , fflush( stdout );
	printf( "In-Core Dimension: %d x %d\n" , globalData.width>>oocLevels , globalData.height>>oocLevels ) , fflush( stdout );
	printf( "       Block Size: %d\n" , blockSize ) , fflush( stdout );

	// Set the local information for the bands
	int start = 0 , index = 0;
	for( int i=0 ; i<_clientCount ; i++ )
	{
		int w = clientSockets[i].clientData.width;
		clientSockets[i].clientData.index = index;
		index += clientSockets[i].clientData.subClients;
		clientSockets[i].clientData.start = start;
		clientSockets[i].clientData.end = clientSockets[i].clientData.cEnd = start+w;
		if( i==_clientCount-1 ) clientSockets[i].clientData.end  = globalData.width;
		start = clientSockets[i].clientData.end;
	
		SendOnSocket( clientSockets[i].client , GetPointer( globalData ) , sizeof(globalData), "Failed to send global data to client" );
		SendOnSocket( clientSockets[i].client , GetPointer( clientSockets[i].clientData ) , sizeof(ClientData) ,"Failed to send client data to client" );
		SendOnSocket( clientSockets[i].client , GetPointer( blockSize ), sizeof( blockSize ) , "Failed to send block size to client" );
	}
	delete[] clientSockets;
	int clipDimensions[2];
	clipDimensions[0] = globalData.cWidth;
	clipDimensions[1] = globalData.cHeight;
	SocketedMultigridServer< Channels , SyncType > server;
	if( !server.SetUp( _address , listenSocket , subClientCount , _iters , _inCoreRes , _minMGRes , _vCycles , _minBandSize , _tileWidth , _tileHeight , _tileExt , _quality , _hdr , _lanes , _verbose , _spherical , _sharpen , _iWeight , _gScale , _blackOut , _showProgress , clipDimensions ) ) return false;
	server.Run();
	return true;
}