/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#ifndef SOCKET_DATA_INCLUDED
#define SOCKET_DATA_INCLUDED

class GlobalData
{
public:
	bool spherical , verbose , sharpen , blackOut , progress , hdr;
	int iters , vCycles , quality , lanes , width , height , cWidth , cHeight , tileWidth , tileHeight;
	double iWeight , gScale;
	char tileExt[16];
};
template<int Channels>
class SolverInfo
{
public:
	double bSquareNorm;
	double rSquareNorm;
	double xSquareNorm;
	double solutionSum[Channels];
};
class ProcessData
{
public:
	int index , offset , maxIndex , maxOffset;
	int start , stop , width , height;
	int startDepth , endDepth;
	int children;
	int depths( void ) { return endDepth - startDepth + 1; }
};
class ProcessSocketData
{
public:
	ProcessSocketData( void ) { left = childX = childB = INVALID_SOCKET; }
	union
	{
		SOCKET left;
		SOCKET right;
	};
	union
	{
		SOCKET childX;
		SOCKET parentX;
	};
	union
	{
		SOCKET childB;
		SOCKET parentB;
	};
};

class ClientData
{
public:
	int index;
	union
	{
		int start;
		int width;
	};
	union
	{
		int end;
		int height;
	};
	union
	{
		int cEnd;
		int subClients;
	};
};
class IPData
{
public:
	IPData( void ) { port = 0; }
	struct in_addr address;
	int port;
};
template< class Data >
class IndexedSocket
{
public:
	SOCKET sock;
	int index;
	Data data;

	static int Sort(const void* v1,const void* v2)
	{
		IndexedSocket *s1 = (IndexedSocket*)v1;
		IndexedSocket *s2 = (IndexedSocket*)v2;
		return s1->index-s2->index;
	}
};


class ProcessConnectionInfo
{
public:
	SOCKET listenLeft , listenChildX , listenChildB;
	DataStream *left, *right;
	SOCKET *childX , *childB , parentX , parentB;
	int children;
	ProcessConnectionInfo( void )
	{
		listenLeft = listenChildX = listenChildB = INVALID_SOCKET;
		left = right = NULL;
		parentX = parentB = INVALID_SOCKET;
		childX = childB = NULL;
		children = 0;
	}
	~ProcessConnectionInfo( void )
	{
		if( childX ) delete[] childX , childX = NULL;
		if( childB ) delete[] childB , childB = NULL;
	}
	void init( int children )
	{
		if( children )
		{
			this->children = children;
			childX = new SOCKET[children];
			childB = new SOCKET[children];
			for( int i=0 ; i<children ; i++ ) childX[i] = childB[i] = INVALID_SOCKET;
		}
	}
};

class AcceptThreadData
{
public:
	int connectCount;
	SOCKET listenSocket,*connectSocket;
	bool success;
};
DWORD WINAPI AcceptThread( LPVOID lpParam )
{
	AcceptThreadData* params = (AcceptThreadData*)lpParam;
	for( int i=0 ; i<params->connectCount ; i++ )
	{
		params->connectSocket[i] = AcceptSocket( params->listenSocket );
		if ( params->connectSocket[i] == INVALID_SOCKET )
		{
			params->success = false;
			delete params;
			return 0;
		}
	}
	CloseSocket( params->listenSocket );
	params->success = true;
	delete params;
	return 0;
}
HANDLE SpawnListenerSocket( SOCKET listenSocket , SOCKET* connectSocket , int connectCount )
{

	HANDLE acceptThread = NULL;
 	AcceptThreadData* params = new AcceptThreadData();	// Have to allocate otherwise memory may be de-allocated (while the thread is running)
	DWORD acceptThreadID;
	params->listenSocket = listenSocket;
	params->connectSocket = connectSocket;
	params->connectCount = connectCount;
	acceptThread = CreateThread
		( 
		NULL,				// default security attributes
		0,					// use default stack size  
		AcceptThread,		// thread function 
		params,				// argument to thread function 
		0,					// use default creation flags 
		&acceptThreadID		// returns the thread identifier
		);
	if( !acceptThread )
	{
		fprintf( stderr , "CreateThread failed: " );
		PrintError();
	}
	return acceptThread;
}

template< int Channels , class Real >
SocketBackedGrid* GetSocketBackedGrid( SOCKET sock , int index , int width , int height , bool blockingSend , bool separateColors )
{
	SendOnSocket( sock , &index , sizeof(index) );
	SendOnSocket( sock , &width , sizeof(width) );
	return new SocketBackedGrid( sock , width * Channels * sizeof(Real) , height , blockingSend , separateColors );
}

template< int Channels , class Real >
MultiSocketBackedGrid< Channels , Real >* GetMultiSocketBackedGrid( SOCKET* sockets , int socketCount , int rows , bool blockingSend , bool separateColors )
{
	IndexedSocket< int >* _sockets = new IndexedSocket< int >[socketCount];
	int *widths = new int[socketCount];

	for( int k=0 ; k<socketCount ; k++ )
	{
		_sockets[k].sock = sockets[k];
		ReceiveOnSocket( _sockets[k].sock , GetPointer( _sockets[k].index ) , sizeof( _sockets[k].index ) );
		ReceiveOnSocket( _sockets[k].sock , GetPointer( _sockets[k].data  ) , sizeof( _sockets[k].data  ) );
	}
	qsort( _sockets , socketCount , sizeof( IndexedSocket<int> ) , IndexedSocket<int>::Sort );
	for( int k=0 ; k<socketCount ; k++ )
	{
		sockets[k] = _sockets[k].sock;
		widths[k] = _sockets[k].data;
	}
	MultiSocketBackedGrid< Channels , Real >* grid = new MultiSocketBackedGrid< Channels , Real >( sockets , widths , socketCount , rows , blockingSend , separateColors );
	delete[] _sockets;
	delete[] widths;
	return grid;
}


#endif // SOCKET_DATA_INCLUDED