/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include <stdio.h>
#include <emmintrin.h>
#include <vector>

#ifdef _WIN64
typedef __int64 signed_size_t;
#else // !_WIN64
typedef   int   signed_size_t;
#endif // _WIN64

template< class C >
class Array
{
	void _assertBounds( signed_size_t idx ) const
	{
		if( idx<min || idx>=max )
		{
			fprintf( stderr , "Array index out-of-bounds: %d <= %d < %d\n" , min , idx , max );
			ASSERT( 0 );
			exit( 0 );
		}
	}
protected:
	bool virt;
	C *data , *_data;
	signed_size_t min , max;
public:
	signed_size_t minimum( void ) const { return min; }
	signed_size_t maximum( void ) const { return max; }

	inline Array( size_t size , size_t alignment , bool clear , bool virt=false , const char* name=NULL )
	{
		this->virt = virt;
		if( virt ) _data = data = ( C* ) VirtualAlloc( NULL , size*sizeof(C) , MEM_RESERVE | MEM_COMMIT , PAGE_READWRITE );
#ifdef WIN32
		else       _data = data = ( C* ) _aligned_malloc( size * sizeof( C ) , alignment );
#else // !WIN32
		else       _data = data = ( C* ) memalign( size * sizeof( C ) , alignment );
#endif // WIN32
		if( clear ) memset( data ,  0 , size * sizeof( C ) );
		else        memset( data , -1 , size * sizeof( C ) );
		min = 0;
		max = size;
	}
	Array( void )
	{
		data = _data = NULL;
		min = max = 0;
	}
	template< class C > static Array FromPointer( C* data , signed_size_t max )
	{
		Array a;
		a._data = NULL;
		a.data = data;
		a.min = 0;
		a.max = max;
		return a;
	}
	template< class C > static Array FromPointer( C* data , signed_size_t min , signed_size_t max )
	{
		Array a;
		a._data = NULL;
		a.data = data;
		a.min = min;
		a.max = max;
		return a;
	}
	inline C& operator[]( signed_size_t idx )
	{
		_assertBounds( idx );
		return data[idx];
	}
	inline const C& operator[]( signed_size_t idx ) const
	{
		_assertBounds( idx );
		return data[idx];
	}
	inline Array operator + ( signed_size_t idx ) const
	{
		Array a;
		a._data = _data;
		a.data = data+idx;
		a.min = min-idx;
		a.max = max-idx;
		return a;
	}
	inline Array  operator +  (        size_t idx ) const { return operator + ( signed_size_t(idx) ); }
#ifdef _WIN64
	inline Array  operator +  (           int idx ) const { return operator + ( signed_size_t(idx) ); }
	inline Array  operator +  (  unsigned int idx ) const { return operator + ( signed_size_t(idx) ); }
#endif _WIN64

	inline Array& operator += ( signed_size_t idx  )
	{
		min -= idx;
		max -= idx;
		data += idx;
		return (*this);
	}
	Array& operator -= ( signed_size_t idx ) { return (*this) += (-idx); }

	Array  operator -  ( signed_size_t idx ) const { return (*this) +  (-idx); }
	Array  operator -  (        size_t idx ) const { return (*this) -  signed_size_t(idx); }
#ifdef _WIN64
	Array  operator -  (           int idx ) const { return (*this) -  signed_size_t(idx); }
	Array  operator -  (  unsigned int idx ) const { return (*this) -  signed_size_t(idx); }
#endif _WIN64

	Array( Array& a )
	{
		if( !a )
		{
			data = _data =  NULL;
			min = max = 0;
		}
		else
		{
			_data = a._data;
			data = a.data;
			min = a.min;
			max = a.max;
		}
	}
	template< class D >
	Array( Array< D >& a )
	{
		_data = NULL;
		if( !a )
		{
			data =  NULL;
			min = max = 0;
		}
		else
		{
			signed_size_t szC = sizeof( C );
			signed_size_t szD = sizeof( D );
			data = (C*)&a[0];
			min = ( a.minimum() * szD ) / szC;
			max = ( a.maximum() * szD ) / szC;
			if( min*szC!=a.minimum()*szD || max*szC!=a.maximum()*szD )
			{
				fprintf( stderr , "Could not convert array [ %d , %d ] * %d => [ %d , %d ] * %d\n" , a.minimum() , a.maximum() , szD , min , max , szC );
				ASSERT( 0 );
				exit( 0 );
			}
		}
	}

	void Free( void )
	{
		if( _data )
		{
			if( virt ) VirtualFree( _data , 0 , MEM_RELEASE );
#ifdef WIN32
			else       _aligned_free( _data );
#else // !WIN32
			else       free( _data );
#endif // WIN32
		}
		(*this) = Array( );
	}
	C* pointer( void ){ return data; }
	bool operator !( void ) { return data==NULL; }
	operator bool( ) { return data!=NULL; }
};

template< class C >
class ConstArray
{
	void _assertBounds( signed_size_t idx ) const
	{
		if( idx<min || idx>=max )
		{
			fprintf( stderr , "Array index out-of-bounds: %d <= %d < %d\n" , min , idx , max );
			ASSERT( 0 );
			exit( 0 );
		}
	}
protected:
	const C *data;
	signed_size_t min , max;
public:
	signed_size_t minimum( void ) const { return min; }
	signed_size_t maximum( void ) const { return max; }

	inline ConstArray( void )
	{
		data = NULL;
		min = max = 0;
	}
	inline ConstArray( Array< C >& a )
	{
		data = a.pointer( );
		min = a.minimum( );
		max = a.maximum( );
	}
	inline ConstArray( ConstArray& a )
	{
		data = a.pointer( );
		min = a.minimum( );
		max = a.maximum( );
	}
	template< class D >
	inline ConstArray( Array< D >& a )
	{
		signed_size_t szC = sizeof( C );
		signed_size_t szD = sizeof( D );
		data = ( const C*)a.pointer( );
		min = ( a.minimum() * szD ) / szC;
		max = ( a.maximum() * szD ) / szC;
		if( min*szC!=a.minimum()*szD || max*szC!=a.maximum()*szD )
		{
			fprintf( stderr , "Could not convert array [ %d , %d ] * %d => [ %d , %d ] * %d\n" , a.minimum() , a.maximum() , szD , min , max , szC );
			ASSERT( 0 );
			exit( 0 );
		}
	}
	template< class D >
	inline ConstArray( ConstArray< D >& a )
	{
		signed_size_t szC = sizeof( C );
		signed_size_t szD = sizeof( D );
		data = ( const C*)a.pointer( );
		min = ( a.minimum() * szD ) / szC;
		max = ( a.maximum() * szD ) / szC;
		if( min*szC!=a.minimum()*szD || max*szC!=a.maximum()*szD )
		{
			fprintf( stderr , "Could not convert array [ %d , %d ] * %d => [ %d , %d ] * %d\n" , a.minimum() , a.maximum() , szD , min , max , szC );
			ASSERT( 0 );
			exit( 0 );
		}
	}
	inline const C& operator[]( signed_size_t idx ) const
	{
		_assertBounds( idx );
		return data[idx];
	}
	inline ConstArray operator + ( signed_size_t idx ) const
	{
		_assertBounds( idx );
		ConstArray a;
		a.data = data+idx;
		a.min = min-idx;
		a.max = max-idx;
		return a;
	}
	ConstArray  operator +  (        size_t idx ) const { return (*this) +  signed_size_t(idx); }
#ifdef _WIN64
	ConstArray  operator +  (           int idx ) const { return (*this) +  signed_size_t(idx); }
	ConstArray  operator +  (  unsigned int idx ) const { return (*this) +  signed_size_t(idx); }
#endif // _WIN64

	inline ConstArray& operator += ( signed_size_t idx  )
	{
		_assertBounds( idx );
		min -= idx;
		max -= idx;
		data += idx;
		return (*this);
	}
	ConstArray& operator -= ( signed_size_t idx )       { return (*this) += (-idx); }

	ConstArray  operator -  ( signed_size_t idx ) const { return (*this) +  (-idx); }
	ConstArray  operator -  (        size_t idx ) const { return (*this) -  signed_size_t(idx); }
#ifdef _WIN64
	ConstArray  operator -  (           int idx ) const { return (*this) -  signed_size_t(idx); }
	ConstArray  operator -  (  unsigned int idx ) const { return (*this) -  signed_size_t(idx); }
#endif _WIN64

	const C* pointer( void ){ return data; }
	bool operator !( void ) { return data==NULL; }
	operator bool( ) { return data!=NULL; }
};

template< class C >
Array< C >& memcpy( Array< C >& destination , const void* source , size_t size )
{
	if( size>destination.maximum()*sizeof(C) )
	{
		fprintf( stderr , "Size of copy exceeds destination maximum: %d > %d\n" , size , destination.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	memcpy( &destination[0] , source , size );
	return destination;
}
template< class C , class D >
Array< C >& memcpy( Array< C >& destination , Array< D >& source , size_t size )
{
	if( size>destination.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of copy exceeds destination maximum: %d > %d\n" , size , destination.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	if( size>source.maximum()*sizeof( D ) )
	{
		fprintf( stderr , "Size of copy exceeds source maximum: %d > %d\n" , size , source.maximum()*sizeof( D ) );
		ASSERT( 0 );
		exit( 0 );
	}
	memcpy( &destination[0] , &source[0] , size );
	return destination;
}
template< class C , class D >
Array< C >& memcpy( Array< C >& destination , ConstArray< D >& source , size_t size )
{
	if( size>destination.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of copy exceeds destination maximum: %d > %d\n" , size , destination.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	if( size>source.maximum()*sizeof( D ) )
	{
		fprintf( stderr , "Size of copy exceeds source maximum: %d > %d\n" , size , source.maximum()*sizeof( D ) );
		ASSERT( 0 );
		exit( 0 );
	}
	memcpy( &destination[0] , &source[0] , size );
	return destination;
}
template< class C , class D >
Array< C >& memcpy( Array< C >& destination , const Array< D >& source , size_t size )
{
	if( size>destination.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of copy exceeds destination maximum: %d > %d\n" , size , destination.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	if( size>source.maximum()*sizeof( D ) )
	{
		fprintf( stderr , "Size of copy exceeds source maximum: %d > %d\n" , size , source.maximum()*sizeof( D ) );
		ASSERT( 0 );
		exit( 0 );
	}
	memcpy( &destination[0] , &source[0] , size );
	return destination;
}
template< class D >
void* memcpy( void* destination , Array< D >& source , size_t size )
{
	if( size>source.maximum()*sizeof( D ) )
	{
		fprintf( stderr , "Size of copy exceeds source maximum: %d > %d\n" , size , source.maximum()*sizeof( D ) );
		ASSERT( 0 );
		exit( 0 );
	}
	memcpy( destination , &source[0] , size );
	return destination;
}
template< class D >
void* memcpy( void* destination , ConstArray< D >& source , size_t size )
{
	if( size>source.maximum()*sizeof( D ) )
	{
		fprintf( stderr , "Size of copy exceeds source maximum: %d > %d\n" , size , source.maximum()*sizeof( D ) );
		ASSERT( 0 );
		exit( 0 );
	}
	memcpy( destination , &source[0] , size );
	return destination;
}
template< class C >
Array< C >& memset( Array< C >& destination , int value , size_t size )
{
	if( size>destination.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of copy exceeds destination maximum: %d > %d\n" , size , destination.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	memset( &destination[0] , value , size );
	return destination;
}

template< class C >
size_t fread( Array< C > destination , size_t eSize , size_t count , FILE* fp )
{
	if( count*eSize>destination.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of read exceeds source maximum: %d > %d\n" , count*eSize , destination.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	return fread( &destination[0] , eSize , count , fp );
}
template< class C >
size_t fwrite( Array< C > source , size_t eSize , size_t count , FILE* fp )
{
	if( count*eSize>source.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of write exceeds source maximum: %d > %d\n" , count*eSize , source.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	return fwrite( &source[0] , eSize , count , fp );
}
template< class C >
size_t fwrite( ConstArray< C > source , size_t eSize , size_t count , FILE* fp )
{
	if( count*eSize>source.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of write exceeds source maximum: %d > %d\n" , count*eSize , source.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	return fwrite( &source[0] , eSize , count , fp );
}
template< class C >
BOOL ReadFile( HANDLE hFile , Array< C > lpBuffer , DWORD nNumberOfBytesToRead , LPDWORD lpNumberOfBytesRead , LPOVERLAPPED lpOverlapped )
{
	if( lpBuffer.maximum()<nNumberOfBytesToRead )
	{
		fprintf( stderr , "Cannot read from file: %d < %d\n" , lpBuffer.maximum() , nNumberOfBytesToRead );
		ASSERT( 0 );
		exit( 0 );
	}
	return ReadFile( hFile , (LPVOID)PointerAddress< C >( lpBuffer ) , nNumberOfBytesToRead , lpNumberOfBytesRead , lpOverlapped );
}
template< class C >
BOOL WriteFile( HANDLE hFile , Array< C > lpBuffer , DWORD nNumberOfBytesToWrite , LPDWORD lpNumberOfBytesWritten , LPOVERLAPPED lpOverlapped )
{
	if( lpBuffer.maximum()<nNumberOfBytesToWrite )
	{
		fprintf( stderr , "Cannot write to file: %d < %d\n" , lpBuffer.maximum() , nNumberOfBytesToWrite );
		ASSERT( 0 );
		exit( 0 );
	}
	return WriteFile( hFile , (LPCVOID)PointerAddress< C >( lpBuffer ) , nNumberOfBytesToWrite , lpNumberOfBytesWritten , lpOverlapped );
}
template< class C >
BOOL WriteFile( HANDLE hFile , ConstArray< C > lpBuffer , DWORD nNumberOfBytesToWrite , LPDWORD lpNumberOfBytesWritten , LPOVERLAPPED lpOverlapped )
{
	if( lpBuffer.maximum()<nNumberOfBytesToWrite )
	{
		fprintf( stderr , "Cannot write to file: %d < %d\n" , lpBuffer.maximum() , nNumberOfBytesToWrite );
		ASSERT( 0 );
		exit( 0 );
	}
	return WriteFile( hFile , (LPCVOID)PointerAddress< C >( lpBuffer ) , nNumberOfBytesToWrite , lpNumberOfBytesWritten , lpOverlapped );
}
template< class C >
size_t recv( SOCKET& s , Array< C > destination , size_t len , int flags )
{
	if( len>destination.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of recv exceeds destination maximum: %d > %d\n" , len , destination.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	return recv( s , (char*)&destination[0] , len , flags );
}
template< class C >
size_t send ( SOCKET s , Array< C > source , size_t len , int flags )
{
	if( len>source.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of send exceeds source maximum: %d > %d\n" , len , source.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	return send( s , (char*)&source[0] , len , flags );
}
template< class C >
size_t send ( SOCKET s , ConstArray< C > source , size_t len , int flags )
{
	if( len>source.maximum()*sizeof( C ) )
	{
		fprintf( stderr , "Size of send exceeds source maximum: %d > %d\n" , len , source.maximum()*sizeof( C ) );
		ASSERT( 0 );
		exit( 0 );
	}
	return send( s , (char*)&source[0] , len , flags );
}
