/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include <windows.h>
#include <atlstr.h>
#include "BaseMultiStreamIO.h"
#include "Socket.h"

//////////////
// IOClient //
//////////////
const int IOClient::BYTES_PER_SECTOR = 1<<9;					// 512B Sector sizes
const int IOClient::IO_BLOCK_SIZE = BYTES_PER_SECTOR<<12;		// 2MB IO Chunks
long long IOClient::ReadBytes  = 0;
long long IOClient::WriteBytes = 0;

IOClient::IOClient( void )
{
	InitializeCriticalSection( &lock );
	server = NULL;
}
IOClient::~IOClient( void )
{
	DeleteCriticalSection( &lock );
}
void IOClient::SetServer( class MultiStreamIOServer* server )
{
	this->server = server;
	if( server ) server->AddClient( this );
}
/////////////////////////
// MultiStreamIOServer //
/////////////////////////
MultiStreamIOServer::MultiStreamIOServer(void)
{
	ioThread = NULL;
	InitializeCriticalSection( &pendingLock );
	InitializeCriticalSection( &clientLock );

	pendingClient = NULL;
	DWORD ioThreadID;
	ioThread = CreateThread( 
		NULL,			// default security attributes
		0,				// use default stack size  
		IOThread,		// thread function 
		this,			// argument to thread function 
		0,				// use default creation flags 
		&ioThreadID);	// returns the thread identifier
	if(!ioThread)
	{
		fprintf(stderr,"Failed to create I/O thread\n");
		exit(0);
	}
}
MultiStreamIOServer::~MultiStreamIOServer(void)
{
	WaitOnIO();
	DeleteCriticalSection( &pendingLock );
	DeleteCriticalSection( &clientLock );
}
bool MultiStreamIOServer::SetPending( IOClient* client )
{
	EnterCriticalSection( &pendingLock );
	if( !pendingClient && client )
	{
		pendingClient = client;
		LeaveCriticalSection( &pendingLock );
		return true;
	}
	else
	{
		LeaveCriticalSection( &pendingLock );
		return false;
	}
}
void MultiStreamIOServer::AddClient( IOClient* client )
{
	EnterCriticalSection( &clientLock );
	clients.push_back( client );
	LeaveCriticalSection( &clientLock );
}
void MultiStreamIOServer::WaitOnIO(void)
{
	if( ioThread )
	{
		while( 1 )
		{
			// As long as the server has work to do, it cannot be terminated.
			EnterCriticalSection( &clientLock );
			if( !clients.size() )
			{
				if( !TerminateThread( ioThread , 0 ) ) fprintf( stderr , "Failed to terminate MultiStreamIOServer thread\n" );
				LeaveCriticalSection( &clientLock );
				break;
			}
			LeaveCriticalSection( &clientLock );
		}

	}
}
int MultiStreamIOServer::clientNum( void )
{
	EnterCriticalSection( &clientLock );
	int sz = clients.size();
	LeaveCriticalSection( &clientLock );
	return sz;
}
DWORD WINAPI MultiStreamIOServer::IOThread( LPVOID lpParam )
{
	MultiStreamIOServer* server = (MultiStreamIOServer*)lpParam;
	std::vector< IOClient* >& clients = server->clients;
	int idx = 0;
	while( 1 )
	{
		EnterCriticalSection( &server->pendingLock );
		if( server->pendingClient )
		{
			if( server->pendingClient->Service() == IOClient::NONE ) Sleep(0);
			else server->pendingClient = NULL;
			LeaveCriticalSection( &server->pendingLock );
		}
		else
		{
			LeaveCriticalSection( &server->pendingLock );
			EnterCriticalSection( &server->clientLock );
			bool ioDone = false;
			for( int i=0 ; i<clients.size() && !ioDone ; i++ )
			{
				idx = (idx+1)%clients.size();
				switch( clients[idx]->Service() )
				{
				case IOClient::COMPLETE:
					clients[idx]->SetServer( NULL );
					clients[idx] = clients[clients.size()-1];
					clients.pop_back();
				case IOClient::SUCCESS:
					ioDone = true;
					break;
				}
			}
			LeaveCriticalSection( &server->clientLock );
			if( !ioDone ) Sleep( 1 );
			else Sleep( 0 );
		}
	}
	return 0;
}
