/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include "Time.h"
#include "GridStream.h"

DWORD MySetFilePointer(HANDLE hFile,LONGLONG distanceToMove)
{
	LARGE_INTEGER li;
	li.QuadPart=distanceToMove;
	DWORD ret=SetFilePointer(hFile,li.LowPart,&li.HighPart,FILE_BEGIN);
	if(ret==INVALID_SET_FILE_POINTER)
	{
		PrintError();
		exit(0);
	}
	return ret;
}
void MyReadFile( HANDLE hFile , Pointer( byte ) lpBuffer , DWORD nNumberOfBytesToRead , LPDWORD lpNumberOfBytesRead , LPOVERLAPPED lpOverlapped )
{
	if( !ReadFile( hFile , lpBuffer , nNumberOfBytesToRead , lpNumberOfBytesRead , lpOverlapped ) )
	{
		fprintf(stderr,"Failed to read file: %d / %d\n",*lpNumberOfBytesRead,nNumberOfBytesToRead);
		PrintError();
		exit(0);
	}
}
void MyWriteFile( HANDLE hFile , Pointer( byte ) lpBuffer , DWORD nNumberOfBytesToWrite , LPDWORD lpNumberOfBytesWritten , LPOVERLAPPED lpOverlapped )
{
	if( !WriteFile( hFile , lpBuffer , nNumberOfBytesToWrite , lpNumberOfBytesWritten , lpOverlapped ) )
	{
		fprintf(stderr,"Failed to write file\n");
		PrintError();
		exit(0);
	}
}
///////////////////////
// NULLStreamingGrid //
///////////////////////
NULLStreamingGrid::NULLStreamingGrid( int rowSize , int rows )
{
	_rows = rows;
	_rowSize = rowSize;
	_data = AllocArray< byte >( rowSize , 1 , "NULLStreamingGrid::_data");
	memset( _data , 0 , rowSize );
}
NULLStreamingGrid::~NULLStreamingGrid( void )
{
	FreeArray( _data );
}

int NULLStreamingGrid::rows( void ) const { return _rows; }
int NULLStreamingGrid::rowSize( void ) const { return _rowSize; }
Pointer( byte ) NULLStreamingGrid::operator [] ( int idx ){ return _data; }

///////////////////////////
// BufferedStreamingGrid //
///////////////////////////
BufferedStreamingGrid::BufferedStreamingGrid(StreamingGrid* sg)
{
	data = NullPointer< byte >( );
	current=0;
	this->sg=sg;
}
BufferedStreamingGrid::~BufferedStreamingGrid(void)
{
	FreeArray( data );
	sg=NULL;
}
int BufferedStreamingGrid::rows		(void) const	{return sg->rows();}
int BufferedStreamingGrid::rowSize	(void) const	{return sg->rowSize();}
void BufferedStreamingGrid::reset	(bool read,int minWindowSize)
{
	this->read=read;
	win=minWindowSize<rows() ? minWindowSize : rows();

	FreeArray( data );
	data = AllocArray< byte >( rowSize() * win , 1 , "BufferedStreamingGrid::_data" );
	if(!data)
	{
		fprintf( stderr , "Failed to allocate memory for BufferedStreamingGrid\n" );
		exit( 0 );
	}
	sg->reset(read,1);
	if(read)
	{
		for(int w=0;w<win;w++)
		{
			Pointer( byte ) row = (*sg)[w];
			memcpy( data+w*rowSize() , row , rowSize() );
			sg->advance();
		}
	}
	current=0;
}
void BufferedStreamingGrid::advance(void)
{
	if(read)
	{
		current++;
		if(current+win-1<rows())
		{
			sg->advance();
			memcpy( data + ((current+win-1)%win)*rowSize() , (*sg)[current+win-1] , rowSize() );
		}
	}
	else
	{
		if(current-win+1>=0)
		{
			memcpy( (*sg)[current-win+1] , data+((current-win+1)%win)*rowSize() , rowSize() );
			sg->advance();
		}
		current++;
	}
}
Pointer( byte ) BufferedStreamingGrid::operator[]	(int idx)
{
	return data+(idx%win)*rowSize();
}
//////////////////////
// MemoryBackedGrid //
//////////////////////
MemoryBackedGrid::MemoryBackedGrid( Pointer( byte ) data , int rs , int r , bool separateColors )
{
	_separateColors = separateColors;
	_del=false;
	_r=r, _rs=rs;
	_data=data;
}
MemoryBackedGrid::MemoryBackedGrid(int rs,int r , bool separateColors )
{
	_separateColors = separateColors;
	_del=true;
	_r=r, _rs=rs;
	_data = AllocArray< byte >( r * rs , 1 , "MemoryBackedGrid::data" );
	if( !_data )	fprintf(stderr,"Failed to allocate memory backed grid of size %d x %d\n" , rs , r ) , exit(0);
}
MemoryBackedGrid::~MemoryBackedGrid(void)
{
	if( _del ) FreeArray( _data );
	_r=_rs=0;
}
int	MemoryBackedGrid::rows(void) const
{
	return _r;
}
int MemoryBackedGrid::rowSize(void) const
{
	return _rs;
}
Pointer( byte ) MemoryBackedGrid::operator[](int idx)
{
	return _data + _rs * idx;
}
//////////////////////
// SocketBackedGrid //
//////////////////////
SocketBackedGrid::SocketBackedGrid( SOCKET sock , int rs , int r , bool bs , bool separateColors )
{
	_separateColors = separateColors;
	_read = true;
	_r = r;
	_rs = rs;
	_sock = sock;
	_bs = bs;
	_data = AllocArray< byte >( rs , 1 , "SocketBackedGrid::data" );
	if( !_data ) fprintfId( stderr , "Failed to allocate\n" ) , exit(0);
}
SocketBackedGrid::~SocketBackedGrid(void)
{
	FreeArray( _data );
	_data = NullPointer< byte >( );
	_r = _rs = 0;
}
int	SocketBackedGrid::rows(void) const { return _r; }
int SocketBackedGrid::rowSize(void) const { return _rs; }
Pointer( byte ) SocketBackedGrid::operator[]( int idx )
{
	if( _read && !_readComplete )
	{
		ReceiveOnSocket( _sock , _data , _rs , _bs , "SocketBackedGrid::operator[]" ) , _readComplete = true;
	}
	return _data;
}
void SocketBackedGrid::advance( void )
{
	if( !_read )
	{
		SendOnSocket( _sock , _data , _rs , _bs , "SocketBackedGrid::advance" );
	}
	else
	{
		if( !_readComplete ) ReceiveOnSocket( _sock , _data , _rs , _bs , "SocketBackedGrid::advance" );
		_readComplete = false;
	}
	_c++;
}
void SocketBackedGrid::reset( bool read , int minWindowSize )
{
	if(!_read ) EndSendOnSocket( _sock , _bs , "SocketBackedGrid::reset" );
	_read = read;
	_readComplete = !_read;
	_c = 0;
	if( _read ) StartReceiveOnSocket( _sock , _bs , "SocketBackedGrid::reset" );
}

//////////////////////
// SharedMemoryGrid // 
//////////////////////
SharedMemoryGrid::SharedMemoryGrid( HANDLE* startHandles , HANDLE* endHandles , int handleCount , Pointer( byte ) row , int rowSize , int rows )
{
	_read = true;
	_r = rows;
	_rs = rowSize;
	_data = row;
	_hCount = handleCount;
	_startHandles = startHandles;
	_endHandles   =   endHandles;
}
SharedMemoryGrid::~SharedMemoryGrid(void)
{
	_data = NullPointer< byte >( );
	_r = _rs = 0;
}
int	SharedMemoryGrid::rows(void) const { return _r; }
int SharedMemoryGrid::rowSize(void) const { return _rs; }
Pointer( byte ) SharedMemoryGrid::operator[]( int idx )
{
	if( !_dataReady )
		MyWaitForMultipleObjects( _hCount , _startHandles , 10000 , "SharedMemoryGrid::operator[%d / %d]" , idx , _r );
	_dataReady = true;
	return _data;
}
void SharedMemoryGrid::advance( void )
{
	for( int i=0 ; i<_hCount ; i++ ) if( !SetEvent( _endHandles[i] ) ) fprintfId( stderr , "SetEvent failed: " ) , PrintError();

	_dataReady = false;
	_c++;
}
void SharedMemoryGrid::reset( bool read , int minWindowSize )
{
	_read = read;
	_dataReady = false;
	_c = 0;
}


////////////////////
// FileBackedGrid //
////////////////////
FileBackedGrid::FileBackedGrid(int rs,int r)
{
	_rSize=_wSize=0;
	_name=_tempnam(".","");
	_fp=fopen(_name,"w+b");
	_read=true;
	_r=r, _rs=rs;
	_data = NullPointer< byte >( );
	_c=_w=0;
}
FileBackedGrid::FileBackedGrid(int rs,int r,FILE* fp)
{
	_rSize=_wSize=0;
	_name=NULL;
	_fp=fp;
	_read=true;
	_r=r, _rs=rs;
	_data = NullPointer< byte >( );
	_c=_w=0;
}
FileBackedGrid::~FileBackedGrid(void)
{
	if(!_read)	for(int i=0;i<_w;i++)	_writeNext();
	FreeArray( _data );
	_r=_rs=_c=_w=0;
	if(_name)
	{
		fclose(_fp);
		remove(_name);
	}
	printf("Read/Write Size: %lld MB/ %lld MB\n",_rSize>>20,_wSize>>20) , fflush( stdout );
}
int	FileBackedGrid::rows(void) const	{	return _r;	}
int FileBackedGrid::rowSize(void) const	{	return _rs;	}
Pointer( byte ) FileBackedGrid::operator[]( int idx )
{
	return _data + _rs * (idx%_w);
}
void FileBackedGrid::reset(bool read,int minWindowSize)
{
	if(!_read)	for(int i=0;i<_w;i++)	_writeNext();
	FreeArray( _data );

	fseek(_fp,0,SEEK_SET);
	_read=read;
	_w=minWindowSize;
	if(_w>_r)	_w=_r;
	_data = AllocArray< byte >( _rs * _w , 1 , "FileBackedGrid::data" );
	if( !_data ) fprintf(stderr,"Failed to allocate in FileBackedGrid::reset\n") , exit(0);
	if(_read)
	{
		_c = -_w;
		for(int i=0;i<_w;i++)	_readNext();
	}
	else	_c = 0;
}
void FileBackedGrid::_readNext(void)
{
	if( _c+_w<_r ) fread( _data + _rs*((_c+_w)%_w) , 1 , _rs , _fp ) , _rSize+=_rs;
	_c++;
}
void FileBackedGrid::_writeNext(void)
{
	_c++;
	if(_c-_w>=0)	fwrite( _data + _rs*((_c+_w)%_w) , 1 , _rs , _fp ) ,	_wSize+=_rs;
}
void FileBackedGrid::advance(void)
{
	if(_read)	_readNext();
	else		_writeNext();
}

//////////////////////////////
// CompressedFileBackedGrid //
//////////////////////////////
CompressedFileBackedGrid::CompressedFileBackedGrid(int rs,int r)
{
	_rSize=_wSize=0;
	_name=_tempnam(".","");
	if(!(_fp=fopen(_name,"wb+")))	fprintf(stderr,"Failed to open: %s\n",_name) , exit(0);
	_read=true;
	_r=r, _rs=rs;
	_c=_w=0;
	_sSize=(_rs+12)*1.002;
	_data = NullPointer< byte >( );
	_scratch = AllocArray< byte >( _sSize , 1 , "CompressedFileBackedGrid::_scratch" );
	if(!_scratch)	fprintf(stderr,"Failed to allocate in CompressedFileBackedGrid::CompressedFileBackedGrid\n") , exit(0);
}
CompressedFileBackedGrid::CompressedFileBackedGrid(int rs,int r,FILE* fp)
{
	_rSize=_wSize=0;
	_name=NULL;
	_fp=fp;
	_read=true;
	_r=r, _rs=rs;
	_c=_w=0;
	_sSize=(_rs+12)*1.002;
	_data = NullPointer< byte >( );
	_scratch = AllocArray< byte >( _sSize , 1 , "CompressedFileBackedGrid::_scratch" );
	if(!_scratch)	fprintf(stderr,"Failed to allocate in CompressedFileBackedGrid::CompressedFileBackedGrid\n") , exit(0);
}
CompressedFileBackedGrid::~CompressedFileBackedGrid(void)
{
	if(!_read)	for(int i=0;i<_w;i++)	_writeNext();
	FreeArray( _data );
	_r=_rs=_c=_w=0;
	if(_name)
	{
		fclose(_fp);
		remove(_name);
		free(_name);
	}
	printf("Read/Write Size: %lld MB / %lld MB\n",_rSize>>20,_wSize>>20) , fflush( stdout );
}
int	CompressedFileBackedGrid::rows(void) const	{	return _r;	}
int CompressedFileBackedGrid::rowSize(void) const	{	return _rs;	}
Pointer( byte ) CompressedFileBackedGrid::operator[](int idx)
{
	return _data + _rs*(idx%_w);
}
void CompressedFileBackedGrid::reset(bool read,int minWindowSize)
{
	if(!_read)	for(int i=0;i<_w;i++)	_writeNext();

	FreeArray( _data );

	if(fseek(_fp,0,SEEK_SET))	fprintf(stderr,"fseek failed in CompressedFileBackedGrid::reset\n") , exit(0);
	_read=read;
	_w=minWindowSize;
	if(_w>_r)	_w=_r;
	_data = AllocArray< byte >( _rs * _w , 1 , "CompressedFileBackedGrid::_data" );
	if(!_data)	fprintf(stderr,"Failed to allocate in CompressedFileBackedGrid::reset\n") , exit(0);
	if(_read)	for( _c = -_w ; _c < 0 ; _readNext() )	;
	else		_c = 0;
}
#include <Util/ZLIB/ZLIB.h>
void CompressedFileBackedGrid::_readNext(void)
{
	if(_c+_w<_r)
	{
		uLong destLen=_rs,sourceLen;
		Bytef *dest = (Bytef*)(LONGLONG(_data)+_rs*((_c+_w)%_w));
		const Bytef *source = (Bytef*)PointerAddress( _scratch );
		if(fread(&sourceLen,sizeof(uLong),1,_fp)!=1)	fprintf(stderr,"Read failed\n") , exit(0);
		if(fread(_scratch,1,sourceLen,_fp)!=sourceLen)	fprintf(stderr,"Read failed\n") , exit(0);
		_rSize+=sizeof(uLong)+sourceLen;

		switch(uncompress(dest,&destLen,source,sourceLen))
		{
		case Z_OK:			break;
		case Z_MEM_ERROR:	fprintf(stderr,"uncompress -- Z_MEM_ERROR\n")  , exit(0);
		case Z_BUF_ERROR:	fprintf(stderr,"uncompress -- Z_BUF_ERROR\n")  , exit(0);
		case Z_DATA_ERROR:	fprintf(stderr,"uncompress -- Z_DATA_ERROR\n") , exit(0);
		}
	}
	_c++;
}
void CompressedFileBackedGrid::_writeNext(void)
{
	_c++;
	if(_c-_w>=0 && _c-_w<_r)
	{
		uLong destLen=_sSize,sourceLen=_rs;
		Bytef *dest = (Bytef*)PointerAddress( _scratch );
		Bytef *source = (Bytef*)(LONGLONG(_data)+_rs*((_c+_w)%_w));
		switch(compress(dest,&destLen,source,sourceLen))
		{
		case Z_OK:	break;
		case Z_MEM_ERROR:	fprintf(stderr,"compress -- Z_MEM_ERROR\n")   , exit(0);
		case Z_BUF_ERROR:	fprintf(stderr,"compress -- Z_BUF_ERROR\n")   , exit(0);
		default:			fprintf(stderr,"compress -- unknown error\n") , exit(0);
		}
		if(fwrite(&destLen,sizeof(uLong),1,_fp)!=1) fprintf(stderr,"Write failed\n") , exit(0);
		if(fwrite(_scratch,1,destLen,_fp)!=destLen) fprintf(stderr,"Write failed\n") , exit(0);
		_wSize+=sizeof(uLong)+destLen;
	}
}
void CompressedFileBackedGrid::advance(void)
{
	if(_read)	_readNext();
	else		_writeNext();
}
