/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include "Time.h"

///////////////////////////
// MultiSocketBackedGrid //
///////////////////////////
template< int Channels , class Real >
MultiSocketBackedGrid< Channels , Real >::MultiSocketBackedGrid( SOCKET* socks , int* rowSizes , int sockCount , int rows , bool blockingSend , bool separateColors )
{
	_read = true;
	_separateColors = separateColors;
	_blockingSend = blockingSend;
	_sockCount = sockCount;
	_rows = rows;
	_rowSize = 0;
	for( int i=0 ; i<_sockCount ; i++ ) _rowSize += rowSizes[i];

	_rowSizes = ( int* )malloc( sizeof(int) * _sockCount );
	_socks = ( SOCKET* )malloc( sizeof(SOCKET) * _sockCount );
	_subData = AllocArray< Pointer( Real ) >( _sockCount , 1 , "MultiSocketBackedGrid::subData" );
	if( !_rowSizes || !_socks || !_subData ) fprintf( stderr , "Failed to allocate\n" ) ,  exit(0);
	for( int i=0 ; i<_sockCount ; i++ )
	{
		_rowSizes[i] = rowSizes[i];
		_socks[i] = socks[i];
		_subData[i] = AllocArray< Real >( Channels * rowSizes[i] , 1 , "MultiSocketBackedGrid::subData[]" );
		if( !_subData[i] ) fprintf( stderr , "Failed to allocate\n" ) , exit(0);
	}

	_data = AllocArray< Real >( _rowSize * Channels , 1 , "MultiSocketBackedGrid::data" );
	if( !_data ) fprintf( stderr , "Failed to allocate\n" ) , exit(0);
}
template< int Channels , class Real >
MultiSocketBackedGrid< Channels , Real >::~MultiSocketBackedGrid(void)
{
	for( int i=0 ; i<_sockCount ; i++ ) FreeArray( _subData[i] );
	FreeArray( _data );
	FreeArray( _subData );
	free( _socks );
	free( _rowSizes );

	_rows = _rowSize = 0;
	_rowSizes = NULL;
	_socks = NULL;
}
template< int Channels , class Real >
int	MultiSocketBackedGrid< Channels , Real >::rows( void ) const { return _rows; }
template< int Channels , class Real >
int MultiSocketBackedGrid< Channels , Real >::rowSize( void ) const { return _rowSize * Channels * sizeof(Real); }
template< int Channels , class Real >
Pointer( byte ) MultiSocketBackedGrid< Channels , Real >::operator[]( int idx )
{
	if( _read && !_readComplete )
	{
		int offset = 0;
		for( int i=0 ; i<_sockCount ; i++ )
		{
			ReceiveOnSocket( _socks[i] , _subData[i] , _rowSizes[i]*Channels*sizeof(Real) , _blockingSend , "MultiSocketBackedGrid<Channels,Real>::[]" );
			if( _separateColors ) for( int c=0 ; c<Channels ; c++ ) memcpy( _data + offset + c*_rowSize , _subData[i]+c*_rowSizes[i] , sizeof(Real)*_rowSizes[i] );
			else memcpy( _data+offset*Channels , _subData[i] , _rowSizes[i]*Channels*sizeof(Real) );
			offset += _rowSizes[i];
		}
		_readComplete = true;
	}
	return ( Pointer( byte ) )_data;
}
template< int Channels , class Real >
void MultiSocketBackedGrid< Channels , Real >::advance( void )
{
	if( !_read )
	{
		int offset = 0;
		for( int i=0 ; i<_sockCount ; i++ )
		{
			if( _separateColors ) for( int c=0 ; c<Channels ; c++ ) memcpy( _subData[i]+c*_rowSizes[i] , _data + offset + c*_rowSize , sizeof(Real)*_rowSizes[i] );
			else memcpy( _subData[i] , _data+offset*Channels , _rowSizes[i]*Channels*sizeof(Real) );
			SendOnSocket( _socks[i] , _subData[i] , _rowSizes[i]*Channels*sizeof(Real) , _blockingSend  , "MultiSocketBackedGrid< Channels,Real>::advance" );
			offset += _rowSizes[i];
		}
	}
	else
	{
		if( !_readComplete )
		{
			int offset = 0;
			for( int i=0 ; i<_sockCount ; i++ )
			{
				if( _separateColors ) for( int c=0 ; c<Channels ; c++ ) memcpy( _subData[i]+c*_rowSizes[i] , _data + offset + c*_rowSize , sizeof(Real)*_rowSizes[i] );
				else memcpy( _subData[i] , _data+offset*Channels , _rowSizes[i]*Channels*sizeof(Real) );
				ReceiveOnSocket( _socks[i] , _subData[i] , _rowSizes[i]*Channels*sizeof(Real) , _blockingSend , "MultiSocketBackedGrid<Channels,Real>::advance"  );
				offset += _rowSizes[i];
			}
		}
		_readComplete = false;
	}
	_current++;
}
template< int Channels , class Real >
void MultiSocketBackedGrid< Channels , Real >::reset( bool read , int minWindowSize )
{
	if(!_read ) for( int i=0 ; i<_sockCount ; i++ ) EndSendOnSocket( _socks[i] , _blockingSend , "MultiSocketBackedGrid<Channels,Real>::reset" );
	_read = read;
	_readComplete = !_read;
	_current = 0;
	if( _read ) for( int i=0 ; i<_sockCount ; i++ ) StartReceiveOnSocket( _socks[i] , _blockingSend , "MultiSocketBackedGrid<Channels,Real >::reset" );
}
